#!/usr/bin/env python

import math
from typing import List, Optional, Tuple, Union, cast

import numpy as np

from ..util.conversion import dB2Linear, linear2dB
from .fading_generators import JakesSampleGenerator, RayleighSampleGenerator

try:
    import matplotlib.pyplot as plt
    # from mpl_toolkits.mplot3d import art3d
    _MATPLOTLIB_AVAILABLE = True
except ModuleNotFoundError:
    _MATPLOTLIB_AVAILABLE = False

Shape = Tuple[int, ...]

# Either a jakes or a Rayleigh model can be used
FadingGenerator = Union[JakesSampleGenerator, RayleighSampleGenerator]

# Type representing something that can be used to index a numpy array
Indexes = Union[np.ndarray, List[int], slice]


class TdlChannelProfile:
    """
    Channel Profile class.

    This class is just a nice way to present known profiles from the norm
    or the literature, which are represented as instances of this class.

    A TDL channel profile store information about the TDL taps. That is, it
    stores the power and delay of each tap. The power and delay of each tap
    can be accessed through the `tap_powers_*` and `tap_delays` properties.

    Some profiles are defined as objects of this class, such as
    ``COST259_TUx``, ``COST259_RAx`` and ``COST259_HTx``. These can be used when
    instantiating a :class:`TdlChannel` object.

    Note that the tap powers and delays are not necessarily `discretized`
    to some sampling interval.

    Parameters
    ----------
    tap_powers_dB : np.ndarray
        The tap powers (in dB). If both `tap_powers_dB` and `tap_delays`
        are None then a single tap with 0dB power will be assumed at delay
        0.
    tap_delays : np.ndarray
        The tap delays.
    name : str
        A name for the channel profile

    Examples
    --------
    >>> jakes_generator = fading_generators.JakesSampleGenerator(Ts=3.25e-8)
    >>> tdlchannel = TdlChannel(jakes_generator, channel_profile=COST259_TUx)
    """
    def __init__(self,
                 tap_powers_dB: Optional[np.ndarray] = None,
                 tap_delays: Optional[np.ndarray] = None,
                 name: str = 'custom') -> None:
        self._name: str = name
        if tap_powers_dB is None and tap_delays is None:
            tap_powers_dB = np.zeros(1)
            tap_delays = np.zeros(1)

        assert (tap_powers_dB is not None)
        assert (tap_delays is not None)

        self._tap_powers_dB: np.ndarray = tap_powers_dB.copy()
        self._tap_powers_dB.flags['WRITEABLE'] = False
        self._tap_powers_linear: np.ndarray = dB2Linear(tap_powers_dB)
        self._tap_powers_linear.flags['WRITEABLE'] = False
        self._tap_delays: np.ndarray = tap_delays.copy()
        self._tap_delays.flags['WRITEABLE'] = False
        self._num_taps: int = tap_delays.size

        self._mean_excess_delay: float \
            = (np.sum(self._tap_powers_linear * self._tap_delays) /
               np.sum(self._tap_powers_linear))

        aux = (np.sum(self._tap_powers_linear * self._tap_delays**2) /
               np.sum(self._tap_powers_linear))
        self._rms_delay_spread = math.sqrt(aux - self._mean_excess_delay**2)

        # Sampling interval when the channel profile is discretized. You
        # can call the
        self._Ts: Optional[float] = None

    # noinspection PyPep8
    @property
    def mean_excess_delay(self) -> float:
        """
        The mean excess delay is the first moment of the power delay profile
        and is defined to be

        .. math:: \\overline{\\tau} = \\frac{\\sum_k P(\\tau_k)\\tau_k}{\\sum_k P(\\tau_k)}

        Returns
        -------
        float
            The mean excess delay.
        """
        # $$\overline{\tau} = \frac{\sum_k P(\tau_k)\tau_k}{\sum_k P(\tau_k)}$$
        return self._mean_excess_delay

    # noinspection PyPep8
    @property
    def rms_delay_spread(self) -> float:
        """
        The RMS delay spread is the square root of the second central moment of
        the power delay profile. It is defined to be

        .. math:: \\sigma_t = \\sqrt{\\overline{t^2} - \\overline{\\tau}^2}

        where

        .. math:: \\overline{\\tau^2}=\\frac{\\sum_k P(\\tau_k)\\tau_k^2}{\\sum_k P(\\tau_k)}

        Typically, when the symbol time period is greater than 10 times
        the RMS delay spread, no ISI equalizer is needed in the receiver.

        Returns
        -------
        float
            The RMS delay spread.
        """
        # $\sigma_t = \sqrt{\overline{t^2} - \overline{\tau}^2}$
        # where
        # $\overline{\tau^2}=\frac{\sum_k P(\tau_k)\tau_k^2}{\sum_k P(\tau_k)}$
        return self._rms_delay_spread

    @property
    def name(self) -> str:
        """
        Get the profile name.

        Returns
        -------
        str
            Profile name.
        """
        return self._name

    @property
    def tap_powers_dB(self) -> np.ndarray:
        """
        Get the tap powers (in dB).

        Returns
        -------
        np.ndarray
            The tap powers (in dB).
        """
        return self._tap_powers_dB

    @property
    def tap_powers_linear(self) -> np.ndarray:
        """
        Get the tap powers (in linear scale).

        Returns
        -------
        np.ndarray
            The tap powers (in linear scale)."""
        return self._tap_powers_linear

    @property
    def tap_delays(self) -> np.ndarray:
        """
        Get the tap delays.

        Returns
        -------
        np.ndarray
            The tap delays.
        """
        return self._tap_delays

    @property
    def num_taps(self) -> int:
        """
        Get the number of taps in the profile.

        Returns
        -------
        int
            Number of taps before discretization (does not count possible
            padding).
        """
        return self._num_taps

    @property
    def num_taps_with_padding(self) -> int:
        """
        Get the number of taps in the profile including zero-padding
        when the profile is discretized.

        If the profile is not discretized an exception is raised.

        Returns
        -------
        int
            Number of taps after discretization (it counts possible any
            added padding).
        """
        if self.Ts is None:
            raise RuntimeError('TdlChannelProfile is not discretized')

        return cast(int, self._tap_delays[-1] + 1)

    @property
    def Ts(self) -> Optional[float]:
        """
        Get the sampling interval used for discretizing this channel
        profile object.

        If it is not discretized then this returns None.

        Returns
        -------
        float, None
            The sampling interval (in seconds).
        """
        return self._Ts

    @property
    def is_discretized(self) -> bool:
        """Returns True if the channel profile is discretized"""
        if self._Ts is None:
            return False

        return True

    def get_discretize_profile(self, Ts: float) -> "TdlChannelProfile":
        """
        Compute the discretized taps (power and delay) and return a new
        discretized TdlChannelProfile object.

        The tap powers and delays of the returned TdlChannelProfile object
        correspond to the taps and delays of the TdlChannelProfile object
        used to call `get_discretize_profile` after discretizing with the
        sampling interval `Ts`.

        Parameters
        ----------
        Ts : float
            The sampling time for the discretization of the tap powers and
            delays.

        Returns
        -------
        TdlChannelProfile
            The discretized channel profile
        """
        if self.is_discretized:
            raise RuntimeError("Trying to discretize a TdlChannelProfile "
                               "object that is already discretized.")

        name = "{0} (discretized)".format(self.name)
        powers, delays = self._calc_discretized_tap_powers_and_delays(Ts)

        discretized_channel_profile = TdlChannelProfile(powers, delays, name)
        discretized_channel_profile._Ts = Ts

        return discretized_channel_profile

    def _calc_discretized_tap_powers_and_delays(
            self, Ts: float) -> Tuple[np.ndarray, np.ndarray]:
        """
        Discretize the taps according to the sampling time.

        The discretized taps will be equally spaced and the delta time from
        two taps corresponds to the sampling time.

        Parameters
        ----------
        Ts : float
            The sampling time.

        Returns
        -------
        np.ndarray, np.ndarray
            A tuple with the discretized powers and delays.
        """
        # Compute delay indices
        delay_indexes, idx_inverse = np.unique(np.round(
            self._tap_delays / Ts).astype(int).flatten(),
                                               return_inverse=True)

        discretized_powers_linear = np.zeros(delay_indexes.size)
        for i, v in enumerate(self.tap_powers_linear):
            discretized_idx = idx_inverse[i]
            discretized_powers_linear[discretized_idx] += v
        discretized_powers_linear /= np.sum(discretized_powers_linear)

        # Compute the discretized powers in dB
        discretized_powers_dB = linear2dB(discretized_powers_linear)

        return discretized_powers_dB, delay_indexes

    def __repr__(self) -> str:  # pragma: no cover
        """Get the representation of the object.

        Returns
        --------
        str
            The object representation as a string.
        """
        return "<TdlChannelProfile: '{0}' ({1} taps)>".format(
            self.name, self.num_taps)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Channel Profile Classes xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# Channel profiles define the power and delays (according to the norm) for
# use with when creating TdlChannel objects
# xxxxxxxxxx Define some known profiles xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# Reference: 3GPP TR 25.943 version 9.0.0 Release 9

# COST 259 Typical Urban
COST259_TUx = TdlChannelProfile(
    np.array([
        -5.7, -7.6, -10.1, -10.2, -10.2, -11.5, -13.4, -16.3, -16.9, -17.1,
        -17.4, -19, -19, -19.8, -21.5, -21.6, -22.1, -22.6, -23.5, -24.3
    ]),
    np.array([
        0, 217, 512, 514, 517, 674, 882, 1230, 1287, 1311, 1349, 1533, 1535,
        1622, 1818, 1836, 1884, 1943, 2048, 2140
    ]) * 1e-9, 'COST259_TU')

# COST 259 Rural Area
COST259_RAx = TdlChannelProfile(
    np.array(
        [-5.2, -6.4, -8.4, -9.3, -10.0, -13.1, -15.3, -18.5, -20.4, -22.4]),
    np.array([0., 42., 101., 129., 149., 245., 312., 410., 469., 528]) * 1e-9,
    'COST259_RA')

# COST 259 Hilly Terrain
COST259_HTx = TdlChannelProfile(
    np.array([
        -3.6, -8.9, -10.2, -11.5, -11.8, -12.7, -13.0, -16.2, -17.3, -17.7,
        -17.6, -22.7, -24.1, -25.8, -25.8, -26.2, -29.0, -29.9, -30.0, -30.7
    ]),
    np.array([
        0., 356., 441., 528., 546., 609., 625., 842., 916., 941., 15000.,
        16172., 16492., 16876., 16882., 16978., 17615., 17827., 17849., 18016.
    ]) * 1e-9, 'COST259_HT')


class TdlImpulseResponse:
    """
    Class that represents impulse response for a TdlChannel object.

    This impulse response corresponds to the generated samples for one or
    more channel realization of the TdlChannel with the configured fading
    generator.

    Parameters
    ----------
    tap_values : np.ndarray
        The tap_values (not including zero padded taps) of a TDL channel
        generated for the non-zero taps. Dimension: `Num sparse taps x
        SHAPE x num_samples`. The value SHAPE here is the shape of the
        fading generator and corresponds to independent impulse
        responses. Often the shape of the used fading generator is None and
        thus the dimension of `tap_values` is just `Num sparse taps x
        num_samples`

    channel_profile : TdlChannelProfile
        The channel profile that was considering to generate this impulse
        response.
    """
    def __init__(self, tap_values: np.ndarray,
                 channel_profile: TdlChannelProfile) -> None:
        assert (isinstance(channel_profile, TdlChannelProfile))
        if channel_profile.Ts is None:
            raise RuntimeError('Channel profile must be discretized')

        self._channel_profile: TdlChannelProfile = channel_profile

        self._tap_values_sparse: np.ndarray = tap_values
        self._tap_values_dense: Optional[
            np.ndarray] = None  # This will be set when needed

    @property
    def tap_values_sparse(self) -> np.ndarray:
        """
        Return the tap values (not including zero padding) as a numpy
        array.

        Returns
        -------
        np.ndarray
            The tap values (not including possible zero padding).
        """
        return self._tap_values_sparse

    @property
    def tap_indexes_sparse(self) -> np.ndarray:
        """
        Return the (sparse) tap indexes.

        Returns
        -------
        The indexes of the non-zero taps.
        """
        return self._channel_profile.tap_delays

    @property
    def Ts(self) -> Optional[float]:
        """
        Return the sampling interval of this impulse response.

        If the impulse response is not discretized this returns None.

        Returns
        -------
        float
            The sampling interval.
        """
        return self._channel_profile.Ts

    @property
    def tap_delays_sparse(self) -> np.ndarray:
        """
        Return the tap delays (which are multiples of the sampling
        interval).

        Returns
        -------
        np.ndarray
            The tap delays.
        """
        return self.tap_indexes_sparse * self.Ts

    @property
    def tap_values(self) -> np.ndarray:
        """
        Return the tap values (including zero padding) as a numpy array.

        Returns
        -------
        np.ndarray
            The tap values (including zero padding).
        """
        if self._tap_values_dense is None:
            self._tap_values_dense = \
                self._get_samples_including_the_extra_zeros()
        return self._tap_values_dense

    @property
    def num_samples(self) -> int:
        """
        Get the number of samples (different, "neighbor" impulse responses)
        stored here.

        Returns
        -------
        int
            The number of samples in the `TdlImpulseResponse` object.
        """
        return cast(int, self._tap_values_sparse.shape[-1])

    @property
    def channel_profile(self) -> TdlChannelProfile:
        """
        Return the channel profile.

        Returns
        -------
        TdlChannelProfile
            The channel profile.
        """
        return self._channel_profile

    def _get_samples_including_the_extra_zeros(self) -> np.ndarray:
        """
        Return the `samples` including the zeros for the zero taps.

        Returns
        -------
        samples_with_zeros : np.ndarray
            The samples including the extra delays containing zeros.
        """
        num_taps_with_padding = self.tap_indexes_sparse[-1] + 1

        # Shape of sparse tap values
        orig_shape = self._tap_values_sparse.shape
        # The first dimension has the sparse taps. This dimension will
        # change to num_taps_with_padding. Note that the last dimension in
        # this shape corresponds to the number of samples.
        new_shape = (num_taps_with_padding, ) + orig_shape[1:]

        samples_with_zeros = np.zeros(new_shape, dtype=complex)

        # Fill the non-zero taps in samples_with_zeros in the correct
        # indexes
        samples_with_zeros[self.tap_indexes_sparse] = \
            self._tap_values_sparse

        # Disable write on this array so that it has no risk of deviating
        # from self._tap_values_sparse
        samples_with_zeros.flags['WRITEABLE'] = False

        return samples_with_zeros

    def get_freq_response(self, fft_size: int) -> np.ndarray:
        """
        Get the frequency response for this impulse response.

        Parameters
        ----------
        fft_size : int
            The size of the FFT to be applied.

        Returns
        -------
        np.ndarray
            The frequency response. Dimension: `fft_size x num_samples`
            for SISO impulse response or `fft_size x num_rx x num_tx x
            num_samples` for MIMO impulse response.
        """
        # Compute the FFT in the "delay" dimension, which captures the
        # multipath characteristics of the channel. The FFT is calculated
        # independently for each column (second dimension), which
        # corresponds to the second dimension is the time dimension (as the
        # channel response changes in time)
        freq_response = np.fft.fft(
            self._get_samples_including_the_extra_zeros(), fft_size, axis=0)
        return freq_response

    def __mul__(self, value: float) -> "TdlImpulseResponse":
        """
        Multiply the impulse response by a float returning a new (scaled)
        impulse response.

        Only the tap values are modified.

        This method is useful for instance to apply the path loss to an
        existing TdlImpulseResponse object.

        Parameters
        ----------
        value : float
            The number to multiply the impulse response object.

        Returns
        -------
        TdlImpulseResponse
            A new (scaled by `value`) TdlImpulseResponse object.
        """
        return TdlImpulseResponse(value * self._tap_values_sparse,
                                  self._channel_profile)

    def __rmul__(self, value: float) -> "TdlImpulseResponse":
        """
        Multiply the impulse response by a float returning a new (scaled)
        impulse response.

        Only the tap values are modified.

        This method is useful for instance to apply the path loss to an
        existing TdlImpulseResponse object.

        Parameters
        ----------
        value : float
            The number to multiply the impulse response object.

        Returns
        -------
        TdlImpulseResponse
            A new (scaled by `value`) TdlImpulseResponse object.
        """
        return self * value

    # noinspection PyUnresolvedReferences
    def plot_impulse_response(self) -> None:  # pragma: no cover
        """
        Plot the impulse response.
        """
        if not _MATPLOTLIB_AVAILABLE:
            raise RuntimeError("Install matplotlib to use this method")

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        num_taps_with_padding = self._channel_profile.num_taps_with_padding
        x = np.arange(num_taps_with_padding)  # 0 -> 66

        for i in range(self.num_samples):
            z = np.abs(self.tap_values[:, i])
            ax.plot(x, [i] * num_taps_with_padding, z)

        # for y_ in y:
        #     z = np.abs(self.tap_values[:, y_])
        #     for x_, z_ in zip(x, z):
        #         line = Line3D(*zip((x_, y_, 0), (x_, y_, z_)),
        #                       marker='o', markevery=(1, 1))
        #         ax.add_line(line)

        ax.set_xlabel('Taps (delay domain)')
        ax.set_ylabel('Time Domain')
        ax.set_zlabel('Channel Amplitude')

        ax.set_xlim3d(0, num_taps_with_padding)
        ax.set_ylim3d(0, self.num_samples)
        ax.set_zlim3d(
            np.abs(self.tap_values).min(),
            np.abs(self.tap_values).max())

        plt.show()

    # noinspection PyUnresolvedReferences
    def plot_frequency_response(self,
                                fft_size: int) -> None:  # pragma: no cover
        """
        Plot the frequency response.

        Parameters
        ----------
        fft_size : int
            The size of the FFT to be applied.
        """
        if not _MATPLOTLIB_AVAILABLE:
            raise RuntimeError("Install matplotlib to use this method")

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')

        x = np.arange(fft_size)
        freq_response = self.get_freq_response(fft_size)

        for i in range(self.num_samples):
            z = np.abs(freq_response[:, i])
            ax.plot(x, [i] * fft_size, z)

        ax.set_xlabel('Taps (delay domain)')
        ax.set_ylabel('Time Domain')
        ax.set_zlabel('Channel Amplitude')

        ax.set_xlim3d(0, fft_size)
        ax.set_ylim3d(0, self.num_samples)
        ax.set_zlim3d(np.abs(freq_response).min(), np.abs(freq_response).max())

        plt.show()

    @staticmethod
    def concatenate_samples(
        impulse_responses: List["TdlImpulseResponse"]
    ) -> "TdlImpulseResponse":
        """
        Concatenate multiple TdlImpulseResponse objects and return the new
        concatenated TdlImpulseResponse.

        This concatenation is performed in the "samples" dimension.

        Parameters
        ----------
        impulse_responses : list[TdlImpulseResponse]
            A list of TdlImpulseResponse objects to be concatenated.

        Returns
        -------
        TdlImpulseResponse
            The new concatenated TdlImpulseResponse.
        """
        num_objs = len(impulse_responses)
        if num_objs < 2:
            if num_objs == 1:
                return impulse_responses[0]

            raise ValueError("impulse_responses must contain "
                             "at least two TdlImpulseResponse "
                             "objects.")

        # We should test if all elements in impulse_responses have
        # the same profile, but in order to avoid too much overhead we only
        # test the first two.
        channel_profile1 = impulse_responses[0].channel_profile
        channel_profile2 = impulse_responses[1].channel_profile
        if channel_profile1 is not channel_profile2:
            raise ValueError("TdlImpulseResponse objects must have the "
                             "same channel profile object")

        tap_values_sparse = np.concatenate(
            [a.tap_values_sparse for a in impulse_responses], axis=-1)

        concatenated_impulse_response = TdlImpulseResponse(
            tap_values_sparse, channel_profile1)

        return concatenated_impulse_response


class TdlChannel:
    """
    Tapped Delay Line channel model, which corresponds to a multipath
    channel.

    You can create a new TdlChannel object either specifying the channel
    profile or specifying both the channel tap powers and delays.

    Parameters
    ----------
    fading_generator : FadingGenerator
        The instance of a fading generator in the `fading_generators`
        module. It should be a subclass of FadingSampleGenerator. The
        fading generator will be used to generate the channel samples.
        If the shape of the fading_generator is not None, then it must
        contain two positive integers, and a MIMO transmission will be
        employed, where the first integer in shape corresponds to the
        number of receive antennas while the second integer corresponds
        to the number of transmit antennas
    channel_profile : TdlChannelProfile
        The channel profile, which specifies the tap powers and delays.
    tap_powers_dB : np.ndarray
        The powers of each tap (in dB). Dimension: `L x 1`
        Note: The power of each tap will be a negative number (in dB).
    tap_delays : np.ndarray
        The delay of each tap (in seconds). Dimension: `L x 1`
    """

    # Note: It would be better to have only the first argument as
    # positional argument and all the others as keyword only arguments. We
    # can do this in Python3 by adding ",*," after the first positional
    # argument thus making all the other arguments keyword only. However,
    # this is not valid in Python2.
    def __init__(self,
                 fading_generator: FadingGenerator,
                 channel_profile: Optional[TdlChannelProfile] = None,
                 tap_powers_dB: Optional[np.ndarray] = None,
                 tap_delays: Optional[np.ndarray] = None,
                 Ts: Optional[float] = None) -> None:
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        if isinstance(fading_generator, JakesSampleGenerator):
            if Ts is None:
                # Ts was not provided, but the fading generator has
                # it. Let's use it then.
                Ts = fading_generator.Ts
            elif Ts != fading_generator.Ts:
                # Ts was provided and the fading generator also has it, but
                # they are not the same value. Let's raise an exception
                raise RuntimeError(
                    "The provided sampling interval Ts is different from "
                    "the one in the Jakes sample generator.")
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # If the user didn't provide the channel profile, but it provided
        # the tap powers and delays, then we use them to create a custom
        # channel profile
        if channel_profile is None:
            # If channel_profile is not provided, then tap_powers_dB and
            # tap_powers_dB must be provided and we will use them to create
            # the channel profile object

            channel_profile = TdlChannelProfile(tap_powers_dB, tap_delays)
        else:
            assert(isinstance(channel_profile, TdlChannelProfile)),\
                ('channel_profile must be an obj '
                 'of the TdlChannelProfile class')
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        # The channel profile is not discretized yet. We need to
        # discretize it.
        if not channel_profile.is_discretized:
            if (isinstance(fading_generator, RayleighSampleGenerator)
                    and Ts is None):
                Ts = 1.0
            # if Ts is None:
            #     raise RuntimeError(
            #         "You must either provide the Ts argument or provide an "
            #         "already discretized TdlChannelProfile object")
            # else:
            #     channel_profile = channel_profile.get_discretize_profile(Ts)
            assert (Ts is not None)
            channel_profile = channel_profile.get_discretize_profile(Ts)
        elif channel_profile.Ts != Ts and Ts is not None:
            # Channel profile is already discretized but it does not agree
            # with the Ts value provided or the one in the fading generator
            raise RuntimeError(
                "Channel profile is already discretized, but it does not "
                "agree with the discretized parameter Ts")
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        # Finally save the channel profile to a member attribute
        self._channel_profile = channel_profile

        shape = fading_generator.shape
        self._fading_generator = fading_generator
        self._set_fading_generator_shape(shape)

        # Last generated impulse response. This will be set when the
        # generate_impulse_response method is called
        self._last_impulse_response: Optional[TdlImpulseResponse] = None

        # If set to True then the channel direction (downlink/uplink) will
        # be reversed. This is only important for MIMO channels.
        self._switched_direction = False

    @property
    def switched_direction(self) -> bool:
        """
        Get the value of `switched_direction`.

        Returns
        -------
        bool
            True if direction is switched and False otherwise.
        """
        return self._switched_direction

    @switched_direction.setter
    def switched_direction(self, value: bool) -> None:
        """
        Set the value of `switched_direction`.

        Parameters
        ----------
        value : bool
            True to switch directions of false to use original direction.
        """
        if not isinstance(value, bool):
            raise TypeError("switched_direction must be a boolean value")
        # TODO: maybe throw a warning if the TdlChannel object is not a
        # MIMO version (has multiple antennas)

        self._switched_direction = value

    def set_num_antennas(self, num_rx_antennas: int,
                         num_tx_antennas: int) -> None:
        """
        Set the number of transmit and receive antennas for MIMO transmission.

        Set both `num_rx_antennas` and `num_tx_antennas` to None for SISO
        transmission

        Parameters
        ----------
        num_rx_antennas : int
            The number of receive antennas.
        num_tx_antennas : int
            The number of transmit antennas.
        """
        self._set_fading_generator_shape((num_rx_antennas, num_tx_antennas))

    def _set_fading_generator_shape(self, new_shape: Optional[Shape]) -> None:
        """
        Set the shape of the fading generator.

        Parameters
        ----------
        new_shape : tuple[int], None
            The new shape of the fading generator. Note that the actual
            shape will be set to (self.num_taps, new_shape)
        """
        if new_shape is None:
            self._fading_generator.shape = (self.num_taps, )
        else:
            # Note that fading_generator.shape must be a tuple
            self._fading_generator.shape = (self.num_taps, ) + new_shape

    @property
    def channel_profile(self) -> TdlChannelProfile:
        """
        Return the channel profile.

        Returns
        -------
        TdlChannelProfile
            The channel profile.
        """
        return self._channel_profile

    @property
    def num_taps(self) -> int:
        """
        Number of taps not including zero taps after discretization.

        Returns
        -------
        int
            The number of taps (not including padding)
        """
        return self._channel_profile.num_taps

    @property
    def num_taps_with_padding(self) -> int:
        """
        Number of taps including zero taps after discretization.

        Returns
        -------
        int
            The number of taps (including padding)
        """
        # This is only valid if _channel_profile is discretized and the
        # tap_delays correspond to integers
        return self._channel_profile.num_taps_with_padding

    def generate_impulse_response(self, num_samples: int = 1) -> None:
        """
        Generate a new impulse response of all discretized taps (not
        including possible zero padding) for `num_samples` channel
        realizations.

        NOTE: This method is automatically called in the `corrupt_data` and
        `corrupt_data_in_freq_domain` methods and you don't need to call it
        befortransmitting data. After one of them has been called the
        generated impulse response can ge get with the
        `get_last_impulse_response` method.

        The number of discretized taps of the generated impulse response will
        depend on the channel delay profile (the tap_delays passed during
        creation of the TdlChannel object) as well as on the sampling
        interval.

        As an example, the COST259 TU channel profile has 20 different taps
        where the last one has a delay equal to 2.14 microseconds. If the
        sampling interval is configured as 3.25e-08 then the discretized
        channel will have more than 60 taps ( including the zeros padding),
        where only 15 taps are different from zero. These 15 taps are what is
        returned by this method.

        Alternatively, with a sampling time of 1e-6 you will end up with
        only 3 discretized taps.

        Parameters
        ----------
        num_samples : int
            The number of samples to generate (for each tap).
        """
        self._fading_generator.generate_more_samples(num_samples)
        channel_samples = self._fading_generator.get_samples()

        # xxxxxxxxxx Apply the power to each tap xxxxxxxxxxxxxxxxxxxxxxxxxx
        # Note that here we only apply the power to the taps. The delays
        # will be applied when the fading is actually used.

        # Note that self._tap_linear_powers_discretized has a single
        # dimension. We need to add singleton dimensions as necessary
        # before we multiply it by channel_samples so that broadcasting
        # works.
        new_shape = [self.num_taps]
        new_shape.extend([1] * (channel_samples.ndim - 1))

        samples = (channel_samples * np.sqrt(
            np.reshape(self._channel_profile.tap_powers_linear[:, np.newaxis],
                       new_shape)))

        impulse_response = TdlImpulseResponse(samples, self._channel_profile)
        self._last_impulse_response = impulse_response

    @property
    def num_tx_antennas(self) -> int:
        """
        Get the number of transmit antennas.

        Returns
        -------
        int
            The number of transmit antennas.
        """
        if self._fading_generator.shape is None or len(
                self._fading_generator.shape) == 1:
            return -1
        return self._fading_generator.shape[2]

    @property
    def num_rx_antennas(self) -> int:
        """
        Get the number of receive antennas.

        Returns
        -------
        int
            The number of receive antennas.
        """
        if self._fading_generator.shape is None or len(
                self._fading_generator.shape) == 1:
            return -1
        return self._fading_generator.shape[1]

    def get_last_impulse_response(self) -> TdlImpulseResponse:
        """
        Get the last generated impulse response.

        A new impulse response is generated when the method `corrupt_data`
        is called. You can use the `get_last_impulse_response` method to
        get the impulse response used to corrupt the last data.

        Returns
        -------
        TdlImpulseResponse
            The impulse response of the channel that was used to corrupt
            the last data.
        """
        if self._last_impulse_response is None:
            raise RuntimeError("No impulse response was generated yet")
        return self._last_impulse_response

    def __prepare_transmit_signal_shape(self,
                                        signal: np.ndarray) -> np.ndarray:
        """
        Helper method called in corrupt_data and corrupt_data_in_freq_domain
        methods to prepare the shape of transmit `signal`.

        It there is only one transmit antenna but signal is 1D, then an
        extra dimension will be added to `signal`. Otherwise the `signal`
        will be just returned.

        Parameters
        ----------
        signal : np.ndarray
            The signal to be transmitted. This should be 1D for SISO
            systems (or SIMO systems) and 2D for MIMO systems.

        Returns
        -------
        np.ndarray
            Either the same signal of signal with an added dimension.
        """
        assert (self._fading_generator.shape is not None)
        if len(self._fading_generator.shape) == 1:
            return signal

        _, num_rx_ant, num_tx_ant = self._fading_generator.shape

        if self.switched_direction:
            # Switched directions
            if num_rx_ant == 1 and signal.ndim == 1:
                signal = np.reshape(signal, (1, signal.size))
        else:
            # Original directions
            if num_tx_ant == 1 and signal.ndim == 1:
                signal = np.reshape(signal, (1, signal.size))
        return signal

    def corrupt_data(self, signal: np.ndarray) -> np.ndarray:
        """
        Transmit the signal though the TDL channel.

        Parameters
        ----------
        signal : np.ndarray
            The signal to be transmitted. This should be 1D for SISO
            systems (or SIMO systems) and 2D for MIMO systems.

        Returns
        -------
        np.ndarray
            The received signal after transmission through the TDL channel
        """
        # Number of symbols to be transmitted
        num_symbols = signal.shape[-1]

        # Prepare the dimension in signal. This will either do nothing
        # (return the same variable), or add an extra singleton dimension
        # if signal is 1D and we have a SIMO system.
        signal = self.__prepare_transmit_signal_shape(signal)

        # Generate an impulse response with `num_symbols` samples that we
        # will use to corrupt the data.
        self.generate_impulse_response(num_symbols)
        assert (self._last_impulse_response is not None)

        # Get the channel memory (number of extra received symbols).
        channel_memory = self.num_taps_with_padding - 1

        # The indexes of the non-zero taps from our impulse response
        tap_indexes_sparse = self._last_impulse_response.tap_indexes_sparse
        # The values of the (sparse) tap
        tap_values_sparse = self._last_impulse_response.tap_values_sparse

        assert (self._fading_generator.shape is not None)
        if len(self._fading_generator.shape) == 1:
            # xxxxxxxxxx SISO Case xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            # The output size will be equal to the number of symbols to transit
            # plus the channel_memory.
            output = np.zeros(num_symbols + channel_memory, dtype=complex)

            for i, d in enumerate(tap_indexes_sparse):
                output[d:d + num_symbols] += tap_values_sparse[i] * signal
            # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        elif len(self._fading_generator.shape) == 3:
            # xxxxxxxxxx MIMO Case xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            # The output size will be equal to the number of symbols to transit
            # plus the channel_memory.
            _, num_rx_ant, num_tx_ant = self._fading_generator.shape

            if self.switched_direction:
                # xxxxxxxxxx Code for reversed direction xxxxxxxxxxxxxxxxxx
                output = np.zeros((num_tx_ant, num_symbols + channel_memory),
                                  dtype=complex)
                for i, d in enumerate(tap_indexes_sparse):
                    for rx_idx in range(num_rx_ant):
                        output[:, d:d + num_symbols] += (
                            tap_values_sparse[i, rx_idx, :, :] *
                            signal[rx_idx])
                # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
            else:
                # xxxxxxxxxx Code for original direction xxxxxxxxxxxxxxxxxx
                output = np.zeros((num_rx_ant, num_symbols + channel_memory),
                                  dtype=complex)

                for i, d in enumerate(tap_indexes_sparse):
                    for tx_idx in range(num_tx_ant):
                        output[:, d:d + num_symbols] += (
                            tap_values_sparse[i, :, tx_idx, :] *
                            signal[tx_idx])
                # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        else:  # pragma: no cover
            raise RuntimeError(
                "Shape of the fading generator of the TdlChannel class must "
                "have either 1 (SISO) or 3 (MIMO) dimensions")

        return output

    def corrupt_data_in_freq_domain(
            self,
            signal: np.ndarray,
            fft_size: int,
            carrier_indexes: Optional[Indexes] = None) -> np.ndarray:
        """
        Transmit the signal through the TDL channel, but in the frequency
        domain.

        This is ROUGHLY equivalent to modulating `signal` with OFDM using
        `fft_size` subcarriers, transmitting through a regular TdlChannel,
        and then demodulating with OFDM to recover the received signal.

        One important difference is that here the channel is considered
        constant during the transmission of `fft_size` elements in
        `signal`, and then it is varied by the equivalent of the variation
        for that number of elements. That is, the channel is block static.

        Parameters
        ----------
        signal : np.ndarray
            The signal to be transmitted. This should be 1D for SISO
            systems (or SIMO systems) and 2D for MIMO systems.
        fft_size : int
            The size of the Fourier transform to get the frequency
            response.
        carrier_indexes : slice | np.ndarray | list[int]
            The indexes of the subcarriers where signal is to be
            transmitted. If it is None assume all subcarriers will be
            used. This can be a slice object or a numpy array of integers.

        Returns
        -------
        np.ndarray
            The received signal after transmission through the TDL channel
        """
        # Number of symbols to be transmitted
        num_symbols = signal.shape[-1]

        # Prepare the dimension in signal. This will either do nothing
        # (return the same variable), or add an extra singleton dimension
        # if signal is 1D and we have a SIMO system.
        signal = self.__prepare_transmit_signal_shape(signal)

        # xxxxxxxxxx Get the block size xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
        if carrier_indexes is None:
            block_size = fft_size
        else:
            # carrier_indexes may be either a slice object of a numpy array
            # of integers with indexes
            if isinstance(carrier_indexes, slice):
                # Get the indexes from the slice object. This is a tuple
                # with (start, stop, step)
                indexes = carrier_indexes.indices(fft_size)
                block_size = (indexes[1] - indexes[0]) // indexes[2]
            else:
                assert isinstance(carrier_indexes, (np.ndarray, list))
                block_size = len(carrier_indexes)

        if num_symbols % block_size != 0:
            raise ValueError("The num of elements in `signal` must be a "
                             "multiple of number of sent elements per "
                             "`fft_size`.")
        # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

        assert (self._fading_generator.shape is not None)

        # Variable to store the impulse responses for each block. We will
        # concatenate these impulse responses at the end so that we can set
        # self._last_impulse_response to the impulse response of all blocks
        impulse_responses = []
        if len(self._fading_generator.shape) == 1:
            # Output variable representing the received signal
            output = np.empty(num_symbols, dtype=complex)
        elif len(self._fading_generator.shape) == 3:
            _, num_rx_ant, num_tx_ant = self._fading_generator.shape
            if self.switched_direction:
                # xxxxxxxxxx Code for reversed direction xxxxxxxxxxxxxxxxxx
                output = np.zeros((num_symbols, num_tx_ant), dtype=complex)
            else:
                # xxxxxxxxxx Code for original direction xxxxxxxxxxxxxxxxxx
                output = np.zeros((num_symbols, num_rx_ant), dtype=complex)
        else:  # pragma: no cover
            raise RuntimeError(
                "Shape of the fading generator of the TdlChannel class must "
                "have either 1 (SISO) or 3 (MIMO) dimensions")

        # Number of full blocks in `signal`
        num_full_blocks = num_symbols // block_size

        for i in range(num_full_blocks):
            start_idx = block_size * i
            end_idx = block_size * (i + 1)

            # Generate next impulse response: the one we will use to
            # transmit the current block (the channel is static during
            # transmission of a single block)
            self.generate_impulse_response(1)
            impulse_responses.append(self.get_last_impulse_response())

            assert (self._last_impulse_response is not None)

            if len(self._fading_generator.shape) == 1:
                # xxxxxxxxxx SISO case xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                # Get the equivalent frequency response of the last
                # generated impulse response. That is what we will use
                # to corrupt the current block of signal
                if carrier_indexes is None:
                    freq_response = \
                        self._last_impulse_response.get_freq_response(
                            fft_size)[:, 0]
                else:
                    freq_response = \
                        self._last_impulse_response.get_freq_response(
                            fft_size)[carrier_indexes, 0]

                output[start_idx:end_idx] = (freq_response *
                                             signal[start_idx:end_idx])
                # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            else:  # len(self._fading_generator.shape) == 3
                # xxxxxxxxxx MIMO Case xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
                # Get the equivalent frequency response of the last
                # generated impulse response. That is what we will use
                # to corrupt the current block of signal
                if carrier_indexes is None:
                    freq_response = \
                        self._last_impulse_response.get_freq_response(
                            fft_size)[:, :, :, 0]
                else:
                    freq_response = \
                        self._last_impulse_response.get_freq_response(
                            fft_size)[carrier_indexes, :, :, 0]

                if self.switched_direction:
                    # xxxxxxxxxx Code for reversed direction xxxxxxxxxxxxxx
                    # noinspection PyUnboundLocalVariable
                    for rx_idx in range(num_rx_ant):
                        output[start_idx:end_idx, :] += (
                            freq_response[:, rx_idx, :] *
                            signal[rx_idx, start_idx:end_idx, np.newaxis])
                else:
                    # xxxxxxxxxx Code for original direction xxxxxxxxxxxxxx
                    # noinspection PyUnboundLocalVariable
                    for tx_idx in range(num_tx_ant):
                        output[start_idx:end_idx, :] += (
                            freq_response[:, :, tx_idx] *
                            signal[tx_idx, start_idx:end_idx, np.newaxis])
                # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

            # Advance the fading generator by "fft_size - 1" to account how
            # much the channel has "changed" during the duration of the
            # current block
            self._fading_generator.skip_samples_for_next_generation(fft_size -
                                                                    1)

        self._last_impulse_response = TdlImpulseResponse.concatenate_samples(
            impulse_responses)

        # Transposition has no effect for the SISO case. For the MIMO case
        # it will make output have dimension `num_rx_ant x num_samples`
        return output.T


class TdlMimoChannel(TdlChannel):
    """
    Tapped Delay Line channel model, which corresponds to a multipath
    channel.

    You can create a new TdlMimoChannel object either specifying the
    channel profile or specifying both the channel tap powers and delays.

    Note that the TdlChannel class can already work with multiple antennas
    if provided `fading_generator` has a shape with two elements (number of
    receive antennas and number of transmit antennas). The TdlMimoChannel
    only adds a slight better interface over TdlChannel class for working
    with MIMO. This class is also useful to test MIMO transmission, with
    the added `num_tx_antennas` and `num_rx_antennas` properties.

    Parameters
    ----------
    fading_generator : FadingGenerator
        The instance of a fading generator in the `fading_generators`
        module.  It should be a subclass of FadingSampleGenerator. The
        fading generator will be used to generate the channel samples.  The
        shape of the fading_generator will be ignored and replaced by
        provided number of antennas.
    channel_profile : TdlChannelProfile
        The channel profile, which specifies the tap powers and delays.
    tap_powers_dB : np.ndarray
        The powers of each tap (in dB). Dimension: `L x 1`
        Note: The power of each tap will be a negative number (in dB).
    tap_delays : np.ndarray
        The delay of each tap (in seconds). Dimension: `L x 1`
    """
    def __init__(self,
                 fading_generator: FadingGenerator,
                 channel_profile: Optional[TdlChannelProfile] = None,
                 tap_powers_dB: Optional[np.ndarray] = None,
                 tap_delays: Optional[np.ndarray] = None,
                 Ts: Optional[float] = None) -> None:
        if fading_generator.shape is None or len(fading_generator.shape) != 2:
            raise RuntimeError(  # pragma: nocover
                "The provided fading_generator for the TdlMimoChannel class"
                " must have a shape with two values")

        super(TdlMimoChannel, self).__init__(fading_generator, channel_profile,
                                             tap_powers_dB, tap_delays, Ts)



#!/usr/bin/env python
"""Implement several metrics for subspaces."""

import math
from typing import cast

import numpy as np

from .projections import calcProjectionMatrix

__all__ = [
    "calc_principal_angles", "calc_chordal_distance_from_principal_angles",
    "calc_chordal_distance", "calc_chordal_distance_2"
]


# TODO: I think calc_principal_angles is not correct when matrix1 e matrix2
# have different sizes. At least obtaining the chordal distance from the
# principal angles does not work when matrix1 and matrix2 have different
# shapes.
def calc_principal_angles(matrix1: np.ndarray,
                          matrix2: np.ndarray) -> np.ndarray:
    """
    Calculates the principal angles between `matrix1` and `matrix2`.

    Parameters
    ----------
    matrix1 : np.ndarray
        A 2D numpy array.
    matrix2 : np.ndarray
        A 2D numpy array.

    Returns
    -------
    np.ndarray
        The principal angles between `matrix1` and `matrix2`. This is a
        1D numpy array.

    See also
    --------
    calc_chordal_distance_from_principal_angles

    Examples
    --------
    >>> A = np.array([[1, 2], [3, 4], [5, 6]])
    >>> B = np.array([[1, 5], [3, 7], [5, -1]])
    >>> print(calc_principal_angles(A, B))
    [0.         0.54312217]
    """
    # First we need to find the orthogonal basis for matrix1 and
    # matrix2. This can be done with the QR decomposition. Note that if
    # matrix1 has 'n' columns then its orthogonal basis is given by the
    # first 'n' columns of the 'Q' matrix from its QR decomposition.
    Q1 = np.linalg.qr(matrix1)[0]
    Q2 = np.linalg.qr(matrix2)[0]

    # TODO: Test who has more columns. Q1 must have dimension grater than
    # or equal to Q2 so that the SVD can be calculated in the order below.
    #
    # See the algorithm in
    # http://sensblogs.wordpress.com/2011/09/07/matlab-codes-for-principal-angles-also-termed-as-canonical-correlation-between-any-arbitrary-subspaces-redirected-from-jen-mei-changs-dissertation/
    S = np.linalg.svd(Q1.conjugate().transpose().dot(Q2),
                      full_matrices=False)[1]

    # The singular values of S vary between 0 and 1, but due to
    # computational impressions there can be some value above 1 (by a very
    # small value). Below we change values greater then 1 to be equal to 1
    # to avoid problems with the arc-cos call later.
    S[S > 1] = 1  # Change values greater then 1 to 1

    # The singular values in the matrix S are equal to the cosine of the
    # principal angles. We can calculate the arc-cosine of each element
    # then.
    return np.arccos(S)


# noinspection PyPep8
def calc_chordal_distance_from_principal_angles(
        principalAngles: np.ndarray) -> float:
    """
    Calculates the chordal distance from the principal angles.

    It is given by the square root of the sum of the squares of the sin of
    the principal angles.

    Parameters
    ----------
    principalAngles : np.ndarray
        Numpy array with the principal angles. This is a 1D numpy array.

    Returns
    -------
    chord_dist : float
        The chordal distance.

    See also
    --------
    calc_principal_angles,
    calc_chordal_distance,
    calc_chordal_distance_2

    Examples
    --------
    >>> A = np.arange(1, 9.)
    >>> A.shape = (4, 2)
    >>> B = np.array([[1.2, 2.1], [2.9, 4.3], [5.2, 6.1], [6.8, 8.1]])
    >>> princ_angles = calc_principal_angles(A, B)
    >>> print(round(calc_chordal_distance_from_principal_angles(princ_angles), 8))
    0.47386786
    """
    # noinspection PyTypeChecker
    summation = (np.sum(np.sin(principalAngles)**2)).item()
    return math.sqrt(summation)


def calc_chordal_distance(matrix1: np.ndarray, matrix2: np.ndarray) -> float:
    """
    Calculates the chordal distance between the two matrices

    Parameters
    ----------
    matrix1 : np.ndarray
        A 2D numpy array.
    matrix2 : np.ndarray
        A 2D numpy array.

    Returns
    -------
    chord_dist : float
        The chordal distance.

    Notes
    -----
    Same as :func:`calc_chordal_distance_2`, but implemented differently.

    See also
    --------
    calc_chordal_distance_2,
    calc_chordal_distance_from_principal_angles

    Examples
    --------
    >>> A = np.arange(1, 9.)
    >>> A.shape = (4, 2)
    >>> B = np.array([[1.2, 2.1], [2.9, 4.3], [5.2, 6.1], [6.8, 8.1]])
    >>> print(round(calc_chordal_distance(A, B), 8))
    0.47386786
    """
    Q1 = np.linalg.qr(matrix1)[0]
    Q2 = np.linalg.qr(matrix2)[0]

    # ncols = matrix1.shape[1]  # Must be equal to matrix2.shape[1].

    # The first ncols columns of Q1 and Q2 make orthogonal basis of
    # ran(matrix1) and ran(matrix2), respectively
    Q1_sqr = Q1.dot(Q1.conjugate().transpose())
    Q2_sqr = Q2.dot(Q2.conjugate().transpose())
    return cast(float, np.linalg.norm(Q1_sqr - Q2_sqr, 'fro') / math.sqrt(2.))


def calc_chordal_distance_2(matrix1: np.ndarray, matrix2: np.ndarray) -> float:
    """
    Calculates the chordal distance between the two matrices

    Parameters
    ----------
    matrix1 : np.ndarray
        A 2D numpy array.
    matrix2 : np.ndarray
        A 2D numpy array.

    Returns
    -------
    chord_dist : float
        The chordal distance.

    Notes
    -----
    Same as :func:`calc_chordal_distance`, but implemented differently.

    See also
    --------
    calc_chordal_distance,
    calc_chordal_distance_from_principal_angles

    Examples
    --------
    >>> A = np.arange(1, 9.)
    >>> A.shape = (4, 2)
    >>> B = np.array([[1.2, 2.1], [2.9, 4.3], [5.2, 6.1], [6.8, 8.1]])
    >>> print(round(calc_chordal_distance_2(A, B), 8))
    0.47386786
    """
    return cast(
        float,  #
        (np.linalg.norm(
            calcProjectionMatrix(matrix1) - calcProjectionMatrix(matrix2),
            'fro') / math.sqrt(2))  #
    )



#!/usr/bin/env python
"""
Module implementing validation functions to define "specs" for configobj
validation.

This module is not intended to be used directly. The functions defined here
are used in the other modules in the :mod:`pyphysim.simulations` package.
"""

import numpy as np
import validate

__all__ = [
    "real_numpy_array_check", "real_scalar_or_real_numpy_array_check",
    "integer_numpy_array_check", "integer_scalar_or_integer_numpy_array_check"
]


def _parse_range_expr(value, converter=float):
    """
    Parse a string in the form of min:max or min:step:max and return a
    numpy array.

    Parameters
    ----------
    value : str
        The string to be parsed.
    converter : callable
        function that converts a string representation to a number.

    Returns
    -------
    np.ndarray
        The parsed numpy array.
    """
    try:
        limits = value.split(':')
        limits = [converter(i) for i in limits]
        if len(limits) == 2:
            value = np.arange(limits[0], limits[1])
        elif len(limits) == 3:
            value = np.arange(limits[0], limits[2], limits[1])
    except Exception:
        raise validate.VdtTypeError(value)

    return value


def _parse_float_range_expr(value):
    """
    Parse a string in the form of min:max or min:step:max and return a
    numpy array (of floats).

    Parameters
    ----------
    value : str
        The string to be parsed.

    Returns
    -------
    np.ndarray
        The parsed numpy array.
    """
    return _parse_range_expr(value, float)


def _parse_int_range_expr(value):
    """
    Parse a string in the form of min:max or min:step:max and return a
    numpy array (of integers).

    Parameters
    ----------
    value : str
        The string to be parsed.

    Returns
    -------
    np.ndarray
        The parsed numpy array.
    """
    return _parse_range_expr(value, int)


# pylint: disable= W0622
# noinspection PyShadowingBuiltins
def real_numpy_array_check(value, min=None, max=None):
    """
    Parse and validate `value` as a numpy array (of floats).

    Value can be either a single number, a range expression in the form of
    min:max or min:step:max, or even a list containing numbers and range
    expressions.

    Parameters
    ----------
    value : str
        The string to be converted. This can be either a single number, a
        range expression in the form of min:max or min:step:max, or even a
        list containing numbers and range expressions.
    min : int
        The minimum allowed value. If the converted value is (or have)
        lower than `min` then the VdtValueTooSmallError exception will be
        raised.
    max : int
        The maximum allowed value. If the converted value is (or have)
        greater than `man` then the VdtValueTooSmallError exception will be
        raised.

    Returns
    -------
    np.ndarray
        The parsed numpy array.

    Notes
    -----
    You can either separate the values with commas or spaces (any comma
    will have the same effect as a space). However, if you separate with
    spaces the values should be in brackets, while if you separate with
    commands there should be no brackets.

    >> SNR = 0,5,10:20
    >> SNR = [0 5 10:20]
    """
    if isinstance(value, str):
        # Remove '[' and ']' if they exist.
        if value[0] == '[' and value[-1] == ']':
            value = value[1:-1].strip()
            value = value.replace(',', ' ')  # Replace commas with spaces
            value = value.split()  # Split based on spaces

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Test if it is a list or not
    if isinstance(value, list):
        # If it is a list, each element can be either a number of a 'range
        # expression' that can be parsed with _parse_float_range_expr. We
        # simple apply real_numpy_array_check on each element in the list
        # to do the work and stack horizontally all the results.
        value = [real_numpy_array_check(a, min, max) for a in value]
        out = np.hstack(value)

    else:
        # It its not a list, it can be either a single number of a 'range
        # expression' that can be parsed with _parse_float_range_expr
        try:
            value = validate.is_float(value)
            out = np.array([value])
        except validate.VdtTypeError:
            out = _parse_float_range_expr(value)

    # xxxxxxxxxx Validate if minimum and maximum allowed values xxxxxxxxxxx
    if min is not None:
        # maybe "min" was passed as a string and thus we need to convert it
        # to a float
        min = float(min)
        if out.min() < min:
            raise validate.VdtValueTooSmallError(out.min())

    if max is not None:
        # maybe "min" was passed as a string and thus we need to convert it
        # to a float
        max = float(max)
        if out.max() > max:
            raise validate.VdtValueTooBigError(out.max())
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    return out


# pylint: disable= W0622
# noinspection PyShadowingBuiltins
def real_scalar_or_real_numpy_array_check(value, min=None, max=None):
    """
    Parse and validate `value` as a float number if possible and, if not,
    parse it as a numpy array (of floats).

    Value can be either a single number, a range expression in the form of
    min:max or min:step:max, or even a list containing numbers and range
    expressions. The difference regarding the `real_numpy_array_check`
    function is that if value is a single number it will be parsed as a
    single float value, instead of being parsed as a real numpy array with
    a single element.

    Parameters
    ----------
    value : str | list[str]
        The string to be converted. This can be either a single number, a
        range expression in the form of min:max or min:step:max, or even a
        list containing numbers and range expressions.
    min : int | float
        The minimum allowed value. If the converted value is (or have)
        lower than `min` then the VdtValueTooSmallError exception will be
        raised.
    max : int | float
        The maximum allowed value. If the converted value is (or have)
        greater than `man` then the VdtValueTooSmallError exception will be
        raised.

    Returns
    -------
    np.ndarray
        The parsed numpy array.

    Notes
    -----
    You can either separate the values with commas or spaces (any comma
    will have the same effect as a space). However, if you separate with
    spaces the values should be in brackets, while if you separate with
    commands there should be no brackets.

    >> SNR = 0,5,10:20
    >> SNR = [0 5 10:20]
    """
    try:
        value = validate.is_float(value, min, max)
    except validate.VdtTypeError:
        value = real_numpy_array_check(value, min, max)

    return value


# noinspection PyShadowingBuiltins
def integer_numpy_array_check(value, min=None, max=None):
    """
    Parse and validate `value` as a numpy array (of integers).

    Value can be either a single number, a range expression in the form of
    min:max or min:step:max, or even a list
    containing numbers and range expressions.

    Parameters
    ----------
    value : str
        The string to be converted. This can be either a single number, a
        range expression in the form of min:max or min:step:max, or even a
        list containing numbers and range expressions.
    min : int
        The minimum allowed value. If the converted value is (or have)
        lower than `min` then the VdtValueTooSmallError exception will be
        raised.
    max : int
        The maximum allowed value. If the converted value is (or have)
        greater than `man` then the VdtValueTooSmallError exception will be
        raised.

    Returns
    -------
    np.ndarray
        The parsed numpy array.

    Notes
    -----
    You can either separate the values with commas or spaces (any comma
    will have the same effect as a space). However, if you separate with
    spaces the values should be brackets, while if you separate with
    commands there should be no brackets.

    >> max_iter = 5,10:20
    >> max_iter = [0 5 10:20]
    """
    if isinstance(value, str):
        # Remove '[' and ']' if they exist.
        if value[0] == '[' and value[-1] == ']':
            value = value[1:-1].strip()
            value = value.replace(',', ' ')  # Replace commas by spaces
            value = value.split()  # Split based on spaces

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Test if it is a list or not
    if isinstance(value, list):
        # If it is a list, each element can be either a number of a
        # 'range expression' that can be parsed with
        # _parse_int_range_expr. We simple apply
        # integer_numpy_array_check on each element in the list to do
        # the work and stack horizontally all the results.
        value = [integer_numpy_array_check(a, min, max) for a in value]
        out = np.hstack(value)

    else:
        # It its not a list, it can be either a single number of a 'range
        # expression' that can be parsed with _parse_int_range_expr
        try:
            value = validate.is_integer(value)
            out = np.array([value])
        except validate.VdtTypeError:
            out = _parse_int_range_expr(value)

    # xxxxxxxxxx Validate if minimum and maximum allowed values xxxxxxxxxxx
    if min is not None:
        # maybe "min" was passed as a string and thus we need to convert it
        # to a integer
        min = int(min)
        if out.min() < min:
            raise validate.VdtValueTooSmallError(out.min())

    if max is not None:
        # maybe "min" was passed as a string and thus we need to convert it
        # to a integer
        max = int(max)
        if out.max() > max:
            raise validate.VdtValueTooBigError(out.max())
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    return out


# noinspection PyShadowingBuiltins
def integer_scalar_or_integer_numpy_array_check(value, min=None, max=None):
    """
    Parse and validate `value` as an integer number if possible and,
    if not, parse it as a numpy array (of integers).

    Value can be either a single number, a range expression in the form of
    min:max or min:step:max, or even a list containing numbers and range
    expressions. The difference regarding the `integer_numpy_array_check`
    function is that if value is a single number it will be parsed as a
    single integer value, instead of being parsed as an integer numpy array
    with a single element.

    Parameters
    ----------
    value : str
        The string to be converted. This can be either a single number, a
        range expression in the form of min:max or min:step:max, or even a
        list containing numbers and range expressions.
    min : int
        The minimum allowed value. If the converted value is (or have)
        lower than `min` then the VdtValueTooSmallError exception will be
        raised.
    max : int
        The maximum allowed value. If the converted value is (or have)
        greater than `man` then the VdtValueTooSmallError exception will be
        raised.

    Returns
    -------
    int | np.ndarray
        The parsed numpy array.

    Notes
    -----
    You can either separate the values with commas or spaces (any comma
    will have the same effect as a space). However, if you separate with
    spaces the values should be brackets, while if you separate with
    commands there should be no brackets.

    >> max_iter = 5,10:20
    >> max_iter = [0 5 10:20]
    """
    try:
        value = validate.is_integer(value, min, max)
    except validate.VdtTypeError:
        value = integer_numpy_array_check(value, min, max)

    return value



#!/usr/bin/env python

from typing import Optional, TypeVar

import numpy as np

from ..util.conversion import dB2Linear

# See http://www.qtc.jp/3GPP/Specs/25996-a00.pdf

NumberOrArray = TypeVar("NumberOrArray", np.ndarray, float)


class AntGainBase:  # pragma: no cover
    """Base class for antenna models.
    """
    def get_antenna_gain(self, angle: NumberOrArray) -> NumberOrArray:
        """
        Get the antenna gain for the given angle.

        Parameters
        ----------
        angle : float | np.ndarray
            Angle between the direction of interest and the boresight of
            the antenna. This can also be a numpy array with angles.

        Returns
        -------
        float | np.ndarray
            The gain (in linear scale) for the provided angle.
        """
        raise NotImplementedError("Implement in a subclass")


class AntGainOmni(AntGainBase):
    """
    Class for Omnidirectional antenna gain model.

    Parameters
    ----------
    ant_gain : float, optional
        The antenna gain (in dBi). If not provided then 0dBi will be assumed.
    """
    def __init__(self, ant_gain: Optional[float] = None):
        super(AntGainOmni, self).__init__()
        self.ant_gain: float
        if ant_gain is None:
            self.ant_gain = 1.0
        else:
            self.ant_gain = dB2Linear(ant_gain)

    def get_antenna_gain(self, angle: NumberOrArray) -> NumberOrArray:
        """
        Get the antenna gain for the given angle.

        Parameters
        ----------
        angle : float | np.ndarray
            Angle between the direction of interest and the boresight of
            the antenna. This can also be a numpy array with angles.

        Returns
        -------
        float | np.ndarray
            The gain (in linear scale) for the provided angle.
        """
        if isinstance(angle, np.ndarray):
            return self.ant_gain * np.ones(angle.shape)  # type: ignore

        return self.ant_gain


class AntGainBS3GPP25996(AntGainBase):
    """
    Class for antenna model defined by 3GPP in the 25996 norm for sectorized
    Base Stations.

    The antenna gain (in dBi) will depend on the number of sectors of the
    Base Station.

    NOTE: The antenna pattern here is targeted for diversity-oriented
    implementations (i.e. large inter-element spacings). For beamforming
    applications that require small spacings, alternative antenna designs
    may have to be considered leading to a different antenna pattern.

    Parameters
    ----------
    number_of_sectors : int
        The number of sectors of the base station. It can be either 3 or 6.
    """
    def __init__(self, number_of_sectors: int = 3):
        super(AntGainBS3GPP25996, self).__init__()

        self.ant_gain: float

        if number_of_sectors == 3:
            self.theta_3db = 70.  # Defined in the norm
            self.Am = 20.  # Maximum attenuation in dB
            self.ant_gain = dB2Linear(14.)  # Antenna gain (in dBi)
        elif number_of_sectors == 6:
            self.theta_3db = 35.  # Defined in the norm
            self.Am = 23.  # Maximum attenuation in dB
            self.ant_gain = dB2Linear(17.)  # Antenna gain (in dBi)
        else:
            raise ValueError(
                "Invalid number of sectors: {0}".format(number_of_sectors))

    # noinspection PyPep8
    def get_antenna_gain(self, angle: NumberOrArray) -> NumberOrArray:
        """
        Get the antenna gain for the given angle.

        Parameters
        ----------
        angle : float | np.ndarray
            Angle between the direction of interest and the boresight of
            the antenna. This can also be a numpy array with angles.

        Returns
        -------
        float | np.ndarray
            The gain (in linear scale) for the provided angle.
        """
        # For the antenna gain model defined in 3GPP 25996 the gain (in dB)
        # is given by
        #\(-\min\left[ 12\left( \frac{\theta}{\theta_{3dB}} \right)^2, A_m \right]\), where \(-180 \geq \theta \geq 180\)
        ant_pattern_gain = dB2Linear(
            -np.minimum(12 * (angle / self.theta_3db)**2, self.Am))
        return self.ant_gain * ant_pattern_gain  # type: ignore



#!/usr/bin/env python
"""Module with Sounding Reference Signal (SRS) related functions"""

from typing import List, Tuple, Union, cast

import numpy as np

from .root_sequence import RootSequence
from .zadoffchu import get_shifted_root_seq

# Type representing a shape
Shape = Tuple[int, ...]

# Type representing something that can be used to index a numpy array
Indexes = Union[np.ndarray, List[int], slice]

__all__ = ['get_srs_seq', 'SrsUeSequence']


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Module Functions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
def get_srs_seq(root_seq: np.ndarray, n_cs: int) -> np.ndarray:
    """
    Get the shifted root sequence suitable as the SRS sequence of a user.

    Parameters
    ----------
    root_seq : np.ndarray
        The root sequence to shift. This is a complex numpy array.
    n_cs : int
        The desired cyclic shift number. This should be an integer from 0
        to 7, where 0 will just return the base sequence, 1 gives the first
        shift, and so on.

    Returns
    -------
    np.ndarray
        The shifted root sequence.

    See Also
    --------
    .get_shifted_root_seq, .dmrs.get_dmrs_seq
    """
    return get_shifted_root_seq(root_seq, n_cs, 8)


# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxx Classes xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
# xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
class UeSequence:
    """
    Reference signal sequence of a single user.

    You should not use this class directly and instead use a class that
    inherits from it and provides the desired reference sequence.

    Parameters
    ----------
    root_seq : RootSequence
        The root sequence of the base station the user is
        associated to. This should be an object of the RootSequence
        class.
    n_cs : int
        The shift index of the user. This can be an integer from 1 to 8.
    user_seq_array : np.ndarray
        The user sequence.
    normalize : bool
        True if the reference signal should be normalized. False otherwise.
    """
    def __init__(self,
                 root_seq: RootSequence,
                 n_cs: int,
                 user_seq_array: np.ndarray,
                 normalize: bool = False):
        self._n_cs = n_cs
        self._root_index = root_seq.index
        self._normalized = normalize

        if normalize is True:
            ndim = user_seq_array.ndim
            if ndim == 1:
                # No cover code -> user_seq_array is a 1D array
                norm_factor = np.linalg.norm(user_seq_array)
            else:
                # Case with Cover code -> user_seq_array first dimension is
                # the cover code dimension, while the second dimension is
                # the sequence elements dimension
                norm_factor = np.linalg.norm(user_seq_array[0])
            self._user_seq_array = \
                user_seq_array / norm_factor
        else:
            self._user_seq_array = user_seq_array

    @property
    def normalized(self) -> bool:
        """True if the reference signal is normalized.
        """
        return self._normalized

    @property
    def size(self) -> int:
        """
        Return the size of the reference signal sequence.

        Returns
        -------
        size : int
            The size of the user's reference signal sequence.

        Examples
        --------
        >>> root_seq1 = RootSequence(root_index=25, Nzc=139)
        >>> user_seq1 = SrsUeSequence(root_seq1, 3)
        >>> user_seq1.size
        139
        >>> root_seq2 = RootSequence(root_index=25, Nzc=139, size=150)
        >>> user_seq2 = SrsUeSequence(root_seq2, 3)
        >>> user_seq2.size
        150
        """
        return cast(int, self.seq_array().size)

    @property
    def shape(self) -> Shape:
        """
        Return the shape of the reference signal sequence.

        Returns
        -------
        Tuple[int]
            The shape of the reference signal sequence.
        """
        return cast(Shape, self.seq_array().shape)

    def seq_array(self) -> np.ndarray:
        """
        Get reference signal sequence as a numpy array.

        Returns
        -------
        seq : np.ndarray
            The user's reference signal sequence.
        """
        return self._user_seq_array

    def __getitem__(self, val: Indexes) -> np.ndarray:
        """
        Index the sequence.

        This will simply return the same indexing of the underlying numpy
        array.

        Parameters
        ----------
        val : any
            Anything accepted as indexing by numpy arrays.
        """
        return self.seq_array()[val]

    def __repr__(self) -> str:  # pragma: no cover
        """
        Get the representation of the object.

        Returns
        -------
        str
            The representation of the object.
        """
        return "<{0}(root_index={1}, n_cs={2})>".format(
            self.__class__.__name__, self._root_index, self._n_cs)

    # xxxxxxxxxx Define some basic methods xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # We can always just get the equivalent numpy array and perform the
    # operations on it, but having these operations defined here is
    # convenient

    # TODO: Make these operation methods (add, mul, etc) also work with
    # RootSequence objects returning a new RootSequence object. Change the
    # docstring type information when you do that.
    def __add__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform addition with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """

        return self.seq_array() + other

    def __radd__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform addition with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() + other

    def __mul__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform multiplication with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() * other

    def __rmul__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform multiplication with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() * other

    def conjugate(self) -> np.ndarray:  # pragma: no cover
        """
        Return the conjugate of the root sequence as a numpy array.

        Returns
        -------
        np.ndarray
            The conjugate of the root sequence.
        """

        return self.seq_array().conj()

    def conj(self) -> np.ndarray:  # pragma: no cover
        """
        Return the conjugate of the root sequence as a numpy array.

        Returns
        -------
        np.ndarray
            The conjugate of the root sequence.
        """

        return self.seq_array().conj()

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


class SrsUeSequence(UeSequence):
    """
    SRS sequence of a single user.

    Parameters
    ----------
    root_seq : RootSequence
        The SRS root sequence of the base station the user is
        associated to. This should be an object of the RootSequence
        class.
    n_cs : int
        The shift index of the user. This can be an integer from 0 to 7.
    normalize : bool
        True if the reference signal should be normalized. False otherwise.
    """
    def __init__(self,
                 root_seq: RootSequence,
                 n_cs: int,
                 normalize: bool = False):
        root_seq_array = root_seq.seq_array()
        user_seq_array = get_srs_seq(root_seq_array, n_cs)
        super(SrsUeSequence, self).__init__(root_seq,
                                            n_cs,
                                            user_seq_array,
                                            normalize=normalize)



#!/usr/bin/env python

# pylint: disable-all
"""
This module contains useful functions to create pgfplots code (latex
code using the pgfplots package) from python data.

One example of tex code for a plot using pgfplots is show below

.. code-block:: latex

    \\begin{tikzpicture}
      \\begin{axis}[axis options]
        \\addplot [plot options]
        plot [options]
        coordinates {
          (0, 0)
          (1, 1)
          (2, 4)
          (3, 9)};
        \\addlegendentry{Legend of the last line};
      \\end{axis}
    \\end{tikzpicture}

"""


def generate_pgfplots_plotline(x, y, errors=None, options=None, legend=None):
    """
    This function generates the code corresponding to the "addplot" command
    in a pgfplots plot for the coordinates given in `x` and `y`.

    If the parameter `errors` is also provided then error bars will be
    added in the `y` direction, while options to the addplot command can be
    passed as a string in the `options` argument.

    Parameters
    ----------
    x : np.ndarray | list[float] | list[int]
        The data for the 'x' axis in the plot.
    y : np.ndarray | list[float] | list[int]
        The data for the 'x' axis in the plot
    errors : np.ndarray | list[float] | list[int], optional
        The error for plotting the errorbars.
    options : str
        pgfplot options for the plot line.
        Ex: "color=red,
        solid,
        mark=square,
        mark options={solid}"
    legend : str
        The legend for the plot line.
    """
    import itertools

    # xxxxxxxxxx Creates the coordinates part of the plot line xxxxxxxxxxxx
    points = zip(x, y)
    num_points = min(len(x), len(y))
    if errors is None:
        points_string = "\n".join([str(p) for p in points])
        plot_line = "plot[]\ncoordinates{{{0}}};".format(points_string)
    else:
        error_points = zip(itertools.repeat(0.0, num_points), errors / 2.0)
        points_and_errors_list = [
            "{0} +- {1}".format(a, b) for a, b in zip(points, error_points)
        ]
        points_string = "\n".join(points_and_errors_list)
        plot_line = ("plot[error bars/.cd, y dir = both, y explicit]\n"
                     "coordinates{{{0}}};").format(points_string)
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    # xxxxxxxxxx Get the whole addplot line xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    if options is None:
        addplot_line = "\\addplot[]\n{0}".format(plot_line)
    else:
        addplot_line = "\\addplot[{1}]\n{0}".format(plot_line, options)
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx

    if legend is not None:
        legend_line = "\n\\addlegendentry{{{0}}};".format(legend)
        addplot_line += legend_line

    return addplot_line
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx



#!/usr/bin/env python
"""Implements a waterfilling method.

The doWF method performs the waterfilling algorithm.
"""

from typing import Tuple

import numpy as np

__all__ = ['doWF']


# noinspection PyUnresolvedReferences
def doWF(vtChannels: np.ndarray,
         dPt: float,
         noiseVar: float = 1.0,
         Es: float = 1.0) -> Tuple[np.ndarray, float]:
    """
    Performs the Waterfilling algorithm and returns the optimum power and
    water level.

    Parameters
    ----------
    vtChannels : np.ndarray
        Numpy array with the channel POWER gains (power of the parallel
        AWGN channels).
    dPt : float
        Total available power.
    noiseVar : float
        Noise variance (power in linear scale).
    Es : float
        Symbol energy (in linear scale).

    Returns
    -------
    (vtOptP, mu) : (np.ndarray, float)
        A tuple with vtOptP and mu, where vtOptP are the optimum powers,
        while mu is the water level.
    """
    # Sort Channels (descending order)
    vtChannelsSortIndexes = np.argsort(vtChannels)[::-1]
    vtChannelsSorted = vtChannels[vtChannelsSortIndexes]
    assert isinstance(vtChannelsSorted, np.ndarray)

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Calculates the water level that touches the worst channel (the higher
    # one) and therefore transmits zero power in this worst channel. After
    # that, calculates the power in each channel (the vector 'Ps') for this
    # water level. If the sum of all of these powers in 'Ps' is less then
    # the total available power, then all we need to do is divide the
    # remaining power equally among all the channels (increase the water
    # level). On the other hand, if the sum of all of these powers in 'Ps'
    # is greater then the total available power then we remove the worst
    # channel and repeat the process.
    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # Calculates minimum water-level $\mu$ required to use all channels
    dNChannels = vtChannels.size
    dRemoveChannels = 0

    minMu = float(noiseVar) / (
        Es * vtChannelsSorted[dNChannels - dRemoveChannels - 1])
    Ps = (minMu - float(noiseVar) /
          (Es * vtChannelsSorted[np.arange(0, dNChannels - dRemoveChannels)]))

    # Ps should be a numpy array
    assert isinstance(Ps, np.ndarray)

    while (sum(Ps) > dPt) and (dRemoveChannels < dNChannels):
        dRemoveChannels += 1
        minMu = float(noiseVar) / (
            Es * vtChannelsSorted[dNChannels - dRemoveChannels - 1])
        Ps = (
            minMu - float(noiseVar) /
            (Es * vtChannelsSorted[np.arange(0, dNChannels - dRemoveChannels)])
        )

    # Distributes the remaining power among the all the remaining channels
    dPdiff = dPt - np.sum(Ps)
    vtOptPaux = dPdiff / (dNChannels - dRemoveChannels) + Ps

    # Put optimum power in the original channel order
    vtOptP = np.zeros([
        vtChannels.size,
    ])
    vtOptP[vtChannelsSortIndexes[np.arange(0, dNChannels -
                                           dRemoveChannels)]] = vtOptPaux
    mu = vtOptPaux[0] + float(noiseVar) / vtChannelsSorted[0]

    return vtOptP, mu



#!/usr/bin/env python
"""
The Pyphysim library provides many useful classes and functions to
perform Monte Carlo simulations of communications systems.
"""

# The short X.Y version.
__version__ = 0.4



#!/usr/bin/env python
"""Module with functions to easily moving data from python to MATLAB."""

import numpy as np

__all__ = ["to_mat_str"]


def to_mat_str(x, format_string='+.12e'):
    """Convert the ndarray 'x' to a string corresponding to the MATLAB
    representation of `x`.

    The to_mat_str function formats numpy arrays of arbitrary dimension in
    a way which can easily copied and pasted into an interactive MATLAB
    session

    Parameters
    ----------
    x : numpy array
        The numpy array to be represented as a MATLAB type.
    format_string : str, optional
        The format_string string to convert each element in `x`.

    Returns
    -------
    converted_string : str
        A string that represents the converted numpy array. You can copy
        this string and past it into a MATLAB session.

    Examples
    --------
    >>> a=np.arange(1,10)
    >>> a.shape=(3,3)
    >>> # Print as a numpy matrix
    >>> print(a)
    [[1 2 3]
     [4 5 6]
     [7 8 9]]
    >>> # Call to_mat_str(a) to print the string representation of the
    >>> # converted matrix
    >>> print(to_mat_str(a))
    [\
+1.000000000000e+00, +2.000000000000e+00, +3.000000000000e+00; \
+4.000000000000e+00, +5.000000000000e+00, +6.000000000000e+00; \
+7.000000000000e+00, +8.000000000000e+00, +9.000000000000e+00]

    """

    # noinspection PyShadowingNames
    def convert_row_or_col(numpy_array, format_string, separator=', '):
        """Convert a one-dimensional numpy array to its MATLAB
        representation

        Parameters
        ----------
        numpy_array : numpy array
            The array to be converted.
        format_string : str
            The format string for the conversion.
        separator : str
            The separator for 2 elements.

        """
        # {0:+.12e}
        # +.12e
        output = []
        if numpy_array.dtype == 'complex':
            format_string = "{{0:{0}}}{{1:{0}}}j".format(
                format_string, format_string)
            for i in numpy_array:
                output.append(format_string.format(i.real, i.imag))
        else:
            format_string = '{{0:{0}}}'.format(format_string)
            for i in numpy_array:
                output.append(format_string.format(i))
        return separator.join(output)

    if x.ndim == 1:
        # 1d input
        output = '[{0}]'.format(convert_row_or_col(x, format_string))
        return output

    if x.ndim == 2:
        if x.shape[1] == 1:
            # This is a Column vector
            output = '[{0}]'.format(
                convert_row_or_col(np.reshape(x, x.size),
                                   format_string,
                                   separator='; '))
        else:
            # This is not a column vector
            output = []
            for row in x:
                output.append(convert_row_or_col(row, format_string))
            output = '[{0}]'.format('; '.join(output))
        return output

    if x.ndim > 2:  # pragma: no cover
        raise NotImplementedError('This case is not implemented')
        # d_to_loop = x.shape[2:]
        # sls = [slice(None, None)] * 2
        # print "reshape([ ",
        # # loop over flat index
        # for i in range(np.prod(d_to_loop)):
        #     # reverse order for matlab
        #     # tricky double reversal to get first index to vary fastest
        #     ind_tuple = np.unravel_index(i, d_to_loop[::-1])[::-1]
        #     ind = sls + list(ind_tuple)
        #     mmat(tx[ind], format_string)

        # print '],[',
        # for i in x.shape:
        #     print '%d' % i,
        # print '])'



#!/usr/bin/env python

# pylint: disable=E1103
from typing import cast

import scipy.constants

from ..util.conversion import linear2dBm


def calc_thermal_noise_power_dBm(T: float, delta_f: float) -> float:
    """
    Calculate the thermal noise power (in dBm) for the given room temperature
    `T` (in Cº) and bandwidth `delta_f` (in Hz).

    Parameters
    ----------
    T : float
        Room temperature in Cesium degrees.
    delta_f : float
        Bandwidth in Hz.

    Returns
    -------
    noise_var : float
        The noise power.
    """
    # Boltzmann constant
    B = cast(float, scipy.constants.Boltzmann)
    K = T + 273.0
    noise_power = B * K * delta_f
    noise_power_dBm = linear2dBm(noise_power)
    return noise_power_dBm



#!/usr/bin/env python
"""Module with Sounding Reference Signal (SRS) related functions"""

from typing import List, Optional, Union, cast

import numpy as np

from .zadoffchu import calcBaseZC, get_extended_ZF

# Type representing something that can be used to index a numpy array
Indexes = Union[np.ndarray, List[int], slice]

__all__ = ['RootSequence']

# List of prime numbers lower than 282.
_SMALL_PRIME_LIST = np.array([
    2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71,
    73, 79, 83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151,
    157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233,
    239, 241, 251, 257, 263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 317,
    331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 389, 397, 401, 409, 419,
    421, 431, 433, 439, 443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 503,
    509, 521, 523, 541, 547, 557, 563, 569, 571, 577, 587, 593, 599, 601, 607,
    613, 617, 619, 631, 641, 643, 647, 653, 659, 661, 673, 677, 683, 691, 701,
    709, 719, 727, 733, 739, 743, 751, 757, 761, 769, 773, 787, 797, 809, 811,
    821, 823, 827, 829, 839, 853, 857, 859, 863, 877, 881, 883, 887, 907, 911,
    919, 929, 937, 941, 947, 953, 967, 971, 977, 983, 991, 997, 1009
])

# Table with root sequence for sequence size equal to 12 (number of
# subcarriers in one PRB)
ROOT_TABLE1 = {
    '0': np.array([-1, 1, 3, -3, 3, 3, 1, 1, 3, 1, -3, 3]),
    '1': np.array([1, 1, 3, 3, 3, -1, 1, -3, -3, 1, -3, 3]),
    '2': np.array([1, 1, -3, -3, -3, -1, -3, -3, 1, -3, 1, -1]),
    '3': np.array([-1, 1, 1, 1, 1, -1, -3, -3, 1, -3, 3, -1]),
    '4': np.array([-1, 3, 1, -1, 1, -1, -3, -1, 1, -1, 1, 3]),
    '5': np.array([1, -3, 3, -1, -1, 1, 1, -1, -1, 3, -3, 1]),
    '6': np.array([-1, 3, -3, -3, -3, 3, 1, -1, 3, 3, -3, 1]),
    '7': np.array([-3, -1, -1, -1, 1, -3, 3, -1, 1, -3, 3, 1]),
    '8': np.array([1, -3, 3, 1, -1, -1, -1, 1, 1, 3, -1, 1]),
    '9': np.array([1, -3, -1, 3, 3, -1, -3, 1, 1, 1, 1, 1]),
    '10': np.array([-1, 3, -1, 1, 1, -3, -3, -1, -3, -3, 3, -1]),
    '11': np.array([3, 1, -1, -1, 3, 3, -3, 1, 3, 1, 3, 3]),
    '12': np.array([1, -3, 1, 1, -3, 1, 1, 1, -3, -3, -3, 1]),
    '13': np.array([3, 3, -3, 3, -3, 1, 1, 3, -1, -3, 3, 3]),
    '14': np.array([-3, 1, -1, -3, -1, 3, 1, 3, 3, 3, -1, 1]),
    '15': np.array([3, -1, 1, -3, -1, -1, 1, 1, 3, 1, -1, -3]),
    '16': np.array([1, 3, 1, -1, 1, 3, 3, 3, -1, -1, 3, -1]),
    '17': np.array([-3, 1, 1, 3, -3, 3, -3, -3, 3, 1, 3, -1]),
    '18': np.array([-3, 3, 1, 1, -3, 1, -3, -3, -1, -1, 1, -3]),
    '19': np.array([-1, 3, 1, 3, 1, -1, -1, 3, -3, -1, -3, -1]),
    '20': np.array([-1, -3, 1, 1, 1, 1, 3, 1, -1, 1, -3, -1]),
    '21': np.array([-1, 3, -1, 1, -3, -3, -3, -3, -3, 1, -1, -3]),
    '22': np.array([1, 1, -3, -3, -3, -3, -1, 3, -3, 1, -3, 3]),
    '23': np.array([1, 1, -1, -3, -1, -3, 1, -1, 1, 3, -1, 1]),
    '24': np.array([1, 1, 3, 1, 3, 3, -1, 1, -1, -3, -3, 1]),
    '25': np.array([1, -3, 3, 3, 1, 3, 3, 1, -3, -1, -1, 3]),
    '26': np.array([1, 3, -3, -3, 3, -3, 1, -1, -1, 3, -1, -3]),
    '27': np.array([-3, -1, -3, -1, -3, 3, 1, -1, 1, 3, -3, -3]),
    '28': np.array([-1, 3, -3, 3, -1, 3, 3, -3, 3, 3, -1, -1]),
    '29': np.array([3, -3, -3, -1, -1, -3, -1, 3, -3, 3, 1, -1])
}

# Table with root sequence for sequence size equal to 24 (number of
# subcarriers in two PRBs)
ROOT_TABLE2 = {
    '0':
    np.array([
        -1, 3, 1, -3, 3, -1, 1, 3, -3, 3, 1, 3, -3, 3, 1, 1, -1, 1, 3, -3, 3,
        -3, -1, -3
    ]),
    '1':
    np.array([
        -3, 3, -3, -3, -3, 1, -3, -3, 3, -1, 1, 1, 1, 3, 1, -1, 3, -3, -3, 1,
        3, 1, 1, -3
    ]),
    '2':
    np.array([
        3, -1, 3, 3, 1, 1, -3, 3, 3, 3, 3, 1, -1, 3, -1, 1, 1, -1, -3, -1, -1,
        1, 3, 3
    ]),
    '3':
    np.array([
        -1, -3, 1, 1, 3, -3, 1, 1, -3, -1, -1, 1, 3, 1, 3, 1, -1, 3, 1, 1, -3,
        -1, -3, -1
    ]),
    '4':
    np.array([
        -1, -1, -1, -3, -3, -1, 1, 1, 3, 3, -1, 3, -1, 1, -1, -3, 1, -1, -3,
        -3, 1, -3, -1, -1
    ]),
    '5':
    np.array([
        -3, 1, 1, 3, -1, 1, 3, 1, -3, 1, -3, 1, 1, -1, -1, 3, -1, -3, 3, -3,
        -3, -3, 1, 1
    ]),
    '6':
    np.array([
        1, 1, -1, -1, 3, -3, -3, 3, -3, 1, -1, -1, 1, -1, 1, 1, -1, -3, -1, 1,
        -1, 3, -1, -3
    ]),
    '7':
    np.array([
        -3, 3, 3, -1, -1, -3, -1, 3, 1, 3, 1, 3, 1, 1, -1, 3, 1, -1, 1, 3, -3,
        -1, -1, 1
    ]),
    '8':
    np.array([
        -3, 1, 3, -3, 1, -1, -3, 3, -3, 3, -1, -1, -1, -1, 1, -3, -3, -3, 1,
        -3, -3, -3, 1, -3
    ]),
    '9':
    np.array([
        1, 1, -3, 3, 3, -1, -3, -1, 3, -3, 3, 3, 3, -1, 1, 1, -3, 1, -1, 1, 1,
        -3, 1, 1
    ]),
    '10':
    np.array([
        -1, 1, -3, -3, 3, -1, 3, -1, -1, -3, -3, -3, -1, -3, -3, 1, -1, 1, 3,
        3, -1, 1, -1, 3
    ]),
    '11':
    np.array([
        1, 3, 3, -3, -3, 1, 3, 1, -1, -3, -3, -3, 3, 3, -3, 3, 3, -1, -3, 3,
        -1, 1, -3, 1
    ]),
    '12':
    np.array([
        1, 3, 3, 1, 1, 1, -1, -1, 1, -3, 3, -1, 1, 1, -3, 3, 3, -1, -3, 3, -3,
        -1, -3, -1
    ]),
    '13':
    np.array([
        3, -1, -1, -1, -1, -3, -1, 3, 3, 1, -1, 1, 3, 3, 3, -1, 1, 1, -3, 1, 3,
        -1, -3, 3
    ]),
    '14':
    np.array([
        -3, -3, 3, 1, 3, 1, -3, 3, 1, 3, 1, 1, 3, 3, -1, -1, -3, 1, -3, -1, 3,
        1, 1, 3
    ]),
    '15':
    np.array([
        -1, -1, 1, -3, 1, 3, -3, 1, -1, -3, -1, 3, 1, 3, 1, -1, -3, -3, -1, -1,
        -3, -3, -3, -1
    ]),
    '16':
    np.array([
        -1, -3, 3, -1, -1, -1, -1, 1, 1, -3, 3, 1, 3, 3, 1, -1, 1, -3, 1, -3,
        1, 1, -3, -1
    ]),
    '17':
    np.array([
        1, 3, -1, 3, 3, -1, -3, 1, -1, -3, 3, 3, 3, -1, 1, 1, 3, -1, -3, -1, 3,
        -1, -1, -1
    ]),
    '18':
    np.array([
        1, 1, 1, 1, 1, -1, 3, -1, -3, 1, 1, 3, -3, 1, -3, -1, 1, 1, -3, -3, 3,
        1, 1, -3
    ]),
    '19':
    np.array([
        1, 3, 3, 1, -1, -3, 3, -1, 3, 3, 3, -3, 1, -1, 1, -1, -3, -1, 1, 3, -1,
        3, -3, -3
    ]),
    '20':
    np.array([
        -1, -3, 3, -3, -3, -3, -1, -1, -3, -1, -3, 3, 1, 3, -3, -1, 3, -1, 1,
        -1, 3, -3, 1, -1
    ]),
    '21':
    np.array([
        -3, -3, 1, 1, -1, 1, -1, 1, -1, 3, 1, -3, -1, 1, -1, 1, -1, -1, 3, 3,
        -3, -1, 1, -3
    ]),
    '22':
    np.array([
        -3, -1, -3, 3, 1, -1, -3, -1, -3, -3, 3, -3, 3, -3, -1, 1, 3, 1, -3, 1,
        3, 3, -1, -3
    ]),
    '23':
    np.array([
        -1, -1, -1, -1, 3, 3, 3, 1, 3, 3, -3, 1, 3, -1, 3, -1, 3, 3, -3, 3, 1,
        -1, 3, 3
    ]),
    '24':
    np.array([
        1, -1, 3, 3, -1, -3, 3, -3, -1, -1, 3, -1, 3, -1, -1, 1, 1, 1, 1, -1,
        -1, -3, -1, 3
    ]),
    '25':
    np.array([
        1, -1, 1, -1, 3, -1, 3, 1, 1, -1, -1, -3, 1, 1, -3, 1, 3, -3, 1, 1, -3,
        -3, -1, -1
    ]),
    '26':
    np.array([
        -3, -1, 1, 3, 1, 1, -3, -1, -1, -3, 3, -3, 3, 1, -3, 3, -3, 1, -1, 1,
        -3, 1, 1, 1
    ]),
    '27':
    np.array([
        -1, -3, 3, 3, 1, 1, 3, -1, -3, -1, -1, -1, 3, 1, -3, -3, -1, 3, -3, -1,
        -3, -1, -3, -1
    ]),
    '28':
    np.array([
        -1, -3, -1, -1, 1, -3, -1, -1, 1, -1, -3, 1, 1, -3, 1, -3, -3, 3, 1, 1,
        -1, 3, -1, -1
    ]),
    '29':
    np.array([
        1, 1, -1, -1, -3, -1, 3, -1, 3, -1, 1, 3, 1, -1, 3, 1, 3, -3, -3, 1,
        -1, -1, 1, 3
    ])
}


class RootSequence:
    """
    Class representing the root sequence of the reference signals.

    The root sequence is generated using two possible formulas, one used
    for a sequence size smaller than :math:`3M_{sc}^{RS}` and another for
    sequence size equal to or greater than :math:`3M_{sc}^{RS}`, where
    :math:`3M_{sc}^{RS}` is the number of subcarriers in a PRB (12
    subcarriers).

    Parameters
    ----------
    root_index : int
        The SRS root sequence index.
    size : int
        The size of the extended Zadoff-Chu sequence. If None then the
        sequence will not be extended and will thus have a size equal
        to Nzc.
    Nzc : int
        The size of the Zadoff-Chu sequence (without any extension). If not
        provided then the largest prime number lower than or equal to
        `size` will be used.
    """
    n_sc_PRB = 12  # Number of subcarriers in a PRB in LTE

    def __init__(self,
                 root_index: int,
                 size: Optional[int] = None,
                 Nzc: Optional[int] = None) -> None:
        if size is None and Nzc is None:
            raise AttributeError("Either 'size' or 'Nzc' (or both) must "
                                 "be provided.")
        if size is None:
            size = Nzc
        assert (isinstance(size, int))

        if Nzc is None:
            Nzc = self._get_largest_prime_lower_than_number(size)

        if size < Nzc:
            raise AttributeError("If 'size' and Nzc are provided, "
                                 "then size must be greater than Nzc")

        self._root_index = root_index
        self._seq_array: np.ndarray = None
        self._extended_seq_array: np.ndarray = None  # Extended Zadoff-Chu sequence

        if size > 2 * self.n_sc_PRB:
            # If size is greater then 2 * n_sc_PRB, the root
            # sequence is an extended Zadoff-Chu sequence. First let's
            # compute the Zadoff-Chu sequence.
            self._seq_array = calcBaseZC(Nzc, root_index)  # Zadoff-Chu seq

            # Now, if size is greater than Nzc, let's compute the cyclic
            # extension of the root sequence.
            if size > Nzc:
                self._extended_seq_array = get_extended_ZF(
                    self._seq_array, size)
        else:  # size must be either n_sc_PRB or 2*n_sc_PRB
            if size == self.n_sc_PRB:
                self._seq_array = np.exp(1j * (np.pi / 4.0) *
                                         ROOT_TABLE1['{0}'.format(root_index)])
            elif size == 2 * self.n_sc_PRB:
                self._seq_array = np.exp(1j * (np.pi / 4.0) *
                                         ROOT_TABLE2['{0}'.format(root_index)])
            else:
                raise AttributeError("Invalid root sequence size")

    @staticmethod
    def _get_largest_prime_lower_than_number(seq_size: int) -> int:
        """
        Get the largest prime number lower than `seq_size`.

        Parameters
        ----------
        seq_size : int
            The sequence size.

        Returns
        -------
        int
            The largest prime number lower than `seq_size`.
        """
        p = _SMALL_PRIME_LIST[_SMALL_PRIME_LIST <= seq_size][-1]
        return int(p)

    @property
    def Nzc(self) -> int:
        """
        Get the size of the Zadoff-Chu sequence (without any extension).

        Returns
        -------
        int
            The value of the Nzc property.
        """
        return cast(int, self._seq_array.size)

    @property
    def size(self) -> int:
        """
        Return the size (with extension) of the sequence.

        If the sequence is not extended than `size()` will return the same
        as `Nzc`.

        Returns
        -------
        size : int
            The size of the extended Zadoff-Chu sequence.

        Examples
        --------
        >>> seq1 = RootSequence(root_index=25, Nzc=139)
        >>> seq1.size
        139
        >>> seq1 = RootSequence(root_index=25, Nzc=139, size=150)
        >>> seq1.size
        150
        """
        if self._extended_seq_array is None:
            return self.Nzc

        return cast(int, self._extended_seq_array.size)

    @property
    def index(self) -> int:
        """
        Return the SRS root sequence index.

        Returns
        -------
        int
            The root sequence index.
        """
        return self._root_index

    def seq_array(self) -> np.ndarray:
        """
        Get the extended Zadoff-Chu root sequence as a numpy array.

        Returns
        -------
        seq : np.ndarray
            The extended Zadoff-Chu sequence
        """
        if self._extended_seq_array is None:
            return self._seq_array

        return self._extended_seq_array

    # xxxxxxxxxx Define some basic methods xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    # We can always just get the equivalent numpy array and perform the
    # operations on it, but having these operations defined here is
    # convenient

    # TODO: Make these operation methods (add, mul, etc) also work with
    # RootSequence objects returning a new RootSequence object. Change the
    # docstring type information when you do that.
    def __add__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform addition with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() + other

    def __radd__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform addition with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() + other

    def __mul__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform multiplication with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() * other

    def __rmul__(self, other: np.ndarray) -> np.ndarray:  # pragma: no cover
        """
        Perform multiplication with `other`.

        Parameters
        ----------
        other : np.ndarray

        Returns
        -------
        np.ndrray
        """
        return self.seq_array() * other

    def __getitem__(self, val: Indexes) -> np.ndarray:
        """
        Index the sequence.

        This will simply return the same indexing of the underlying numpy
        array.

        Parameters
        ----------
        val : Indexes
            Anything accepted as indexing by numpy arrays.

        Returns
        -------
        np.ndarray
        """
        return self.seq_array()[val]

    def conjugate(self) -> np.ndarray:  # pragma: no cover
        """
        Return the conjugate of the root sequence as a numpy array.

        Returns
        -------
        np.ndarray
            The conjugate of the root sequence.
        """
        return self.seq_array().conj()

    def conj(self) -> np.ndarray:  # pragma: no cover
        """
        Return the conjugate of the root sequence as a numpy array.

        Returns
        -------
        np.ndarray
            The conjugate of the root sequence.
        """

        return self.seq_array().conj()

    def __repr__(self) -> str:  # pragma: no cover
        """
        Get the representation of the object.

        Returns
        -------
        str
            The representation of the object.
        """
        if self._extended_seq_array is None:
            return ("<SrsRootSequence("
                    "root_index={0},Nzc={1})>").format(self._root_index,
                                                       self._seq_array.size)

        return ("<SrsRootSequence("
                "root_index={0},size={2},Nzc={1})>").format(
                    self._root_index, self._seq_array.size,
                    self._extended_seq_array.size)

    # xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx



#!/usr/bin/env python

# pylint: disable=E1103
import math
from typing import Any, Optional, Tuple, Union

import numpy as np

from ..util.misc import randn_c

Shape = Tuple[int, ...]


# noinspection PyPep8
def generate_jakes_samples(
        Fd: float,
        Ts: float = 1e-3,
        NSamples: int = 100,
        L: int = 8,
        shape: Optional[Shape] = None,
        current_time: float = 0,
        phi_l: Optional[np.ndarray] = None,
        psi_l: Optional[np.ndarray] = None) -> Tuple[float, np.ndarray]:
    """
    Generates channel samples according to the Jakes model.

    This functions generates channel samples for a single tap according to
    the Jakes model given by

    .. math::
       :label: jakes_model

       h(t) = \\frac{1}{\\sqrt{L}}\\sum_{l=0}^{L-1}\\exp\\{j[2\\pi f_D \\cos(\\phi_l)t+\\psi_l]\\}

    Parameters
    ----------
    Fd : float
        The Doppler frequency (in Hertz).
    Ts : float
        The sample interval (in seconds).
    NSamples : int
        The number of samples to generate.
    L : int
        The number of rays for the Jakes model.
    shape : tuple[int]
        The shape of the generated channel. This is used to generate MIMO
        channels. For instance, in order to generate channels samples for a
        MIMO scenario with 3 receive antennas and 2 transmit antennas use a
        shape of (3, 2).
    current_time : float
        The current start time
    phi_l : np.ndarray
        The "phi" part in Jakes model
    psi_l : np.ndarray
        The "psi" part in Jakes model

    Returns
    -------
    (float, np.ndarray)
        The **first element** in the returned tuple is the **new current
        time** (that should be used the next time this function is called to
        'continue' the fading).

        The **second** element in the returned tuple is the **generated
        channel**. If `shape` is None the the shape of the returned h is
        equal to ( NSamples,). That is, h is a 1-dimensional numpy array. If
        `shape` was provided then the shape of h is the provided shape with
        an additional dimension for the time (the last dimension). For
        instance, if a `shape` of (3, 2) was provided then the shape of the
        returned h will be (3, 2, NSamples).
    """
    # Generate time samples
    t = np.arange(
        current_time,  # Start time
        NSamples * Ts + current_time,
        Ts * 1.0000000001)

    if phi_l is None:
        if shape is None:
            phi_l = np.random.rand(L, 1)
        else:
            phi_l = np.random.rand(L, *shape, 1)

    if psi_l is None:
        if shape is None:
            psi_l = np.random.rand(L, 1)
        else:
            psi_l = np.random.rand(L, *shape, 1)

    # Update the self._current_time variable with the value of the next
    # time sample that should be generated when _generate_time_samples
    # is called again.
    new_current_time = t[-1] + Ts

    h = (math.sqrt(1.0 / L) * np.sum(
        np.exp(1j * (2 * np.pi * Fd * np.cos(phi_l) * t + psi_l)), axis=0))

    return new_current_time, h


class FadingSampleGenerator:
    """
    Base class for fading generators.

    Parameters
    ----------
    shape : tuple[int] | int, optional
        The shape of the sample generator. Each time
        `generate_more_samples(num_samples)` method is called it will
        generate samples with this shape as the first dimensions.
    """
    def __init__(self, shape: Optional[Union[int, Shape]] = None) -> None:
        self._shape: Optional[Shape]
        # Call the setter of the shape property to avoid duplicating the code
        # here
        FadingSampleGenerator.shape.fset(self, shape)  # type: ignore

        # Set this variable in a derived class with the next samples
        # everytime the generate_more_samples method is called. Note that
        # generate_more_samples should take the value of self._shape into
        # account.
        self._samples: Optional[np.ndarray] = None

    @property
    def shape(self) -> Optional[Shape]:
        """
        Get the shape of the sampling generator

        This is the shape of the samples that will be generated (not
        including num_samples).

        Returns
        -------
        tuple[int] | None
        """
        return self._shape

    @shape.setter
    def shape(self, new_shape: Optional[Union[int, Shape]]) -> None:
        """
        Set the shape of the sampling generator.

        This is the shape of the samples that will be generated (not
        including num_samples).

        Parameters
        ----------
        new_shape : None | int | tuple[int]
            The shape of the generated channel.
        """
        if isinstance(new_shape, int):
            self._shape = (new_shape, )
        else:
            self._shape = new_shape

    def get_samples(self) -> np.ndarray:
        """
        Get the last generated sample.

        Returns
        -------
        np.ndarray
        """
        return self._samples

    def generate_more_samples(
            self,
            num_samples: Optional[int] = None) -> None:  # pragma: nocover
        """
        Generate next samples.

        When implementing this method in a subclass you must take the value
        of the self._shape attribute into account.

        Parameters
        ----------
        num_samples : int, optional
            Number of samples (with the provided shape) to generate. If not
            provided it will be assumed to be 1.
        """
        raise NotImplementedError("Implement in a subclass")

    def skip_samples_for_next_generation(
            self, num_samples: int) -> None:  # pragma: no cover
        """
        Advance sample generation process by `num_samples` similarly to
        what would happen if you call `generate_more_samples(
        num_samples=num_samples)`, but without actually generating the
        samples.

        Parameters
        ----------
        num_samples : int
            How many samples to skip.
        """
        raise NotImplementedError("Implement in a subclass")

    def get_similar_fading_generator(self) -> Any:  # pragma: no cover
        """
        Get a similar fading generator with the same configuration,
        but that generates independent samples.
        """
        # Note: Don't forget to copy self._shape in subclasses, besides any
        # member attribute in the subclass
        raise NotImplementedError("Implement in a subclass")


class RayleighSampleGenerator(FadingSampleGenerator):
    """
    Class that generates fading samples from a Raleigh distribution.

    Parameters
    ----------
    shape : int | tuple[int] | None
        The shape of the sample generator. Each time the
        `generate_jakes_samples` method is called it will generate samples
        with this shape. If not provided, then 1 will be assumed.
    """
    def __init__(self, shape: Optional[Union[int, Shape]] = None) -> None:
        super(RayleighSampleGenerator, self).__init__(shape)

        # Generate first sample
        self.generate_more_samples()

    def generate_more_samples(self, num_samples: Optional[int] = None) -> None:
        """
        Generate next samples.

        Parameters
        ----------
        num_samples : int, optional
            Number of samples (with the provided shape) to generate. If not
            provided it will be assumed to be 1.
        """
        shape = self.shape

        if num_samples is None:
            if shape is None:
                self._samples = randn_c()
            else:
                # noinspection PyArgumentList
                self._samples = randn_c(*shape)
        elif self.shape is None:
            self._samples = randn_c(num_samples)
        else:
            assert (shape is not None)
            new_shape = list(shape)
            new_shape.append(num_samples)
            self._samples = randn_c(*new_shape)

    def skip_samples_for_next_generation(
            self, num_samples: int) -> None:  # pragma: no cover
        """
        Advance sample generation process by `num_samples` similarly to
        what would happen if you call `generate_more_samples(
        num_samples=num_samples)`, but without actually generating the
        samples.

        Since the samples generated by RayleighSampleGenerator are
        independent, calling this method has no effect.

        Parameters
        ----------
        num_samples : int
            How many samples to skip. This is ignored in the
            RayleighSampleGenerator. Since the different samples are
            uncorrelated then calling `skip_samples_for_next_generation`
            does not do anything.
        """

    def get_similar_fading_generator(self) -> Any:
        """
        Get a similar fading generator with the same configuration,
        but that generates independent samples.

        Returns
        -------
        RayleighSampleGenerator
            Another RayleighSampleGenerator object with the same
            configuration of this object.
        """
        return RayleighSampleGenerator(self._shape)


# TODO: Remove the RS parameter or make it work with the
# get_similar_fading_generator method.  You could also move it to the base
# class and add it as an argument to RayleighSampleGenerator
# noinspection PyPep8
class JakesSampleGenerator(FadingSampleGenerator):
    """
    Class that generated fading samples according to the Jakes model given
    by

    .. math:: h(t) = \\frac{1}{\\sqrt{L}}\\sum_{l=0}^{L-1}\\exp\\{j[2\\pi f_D \\cos(\\phi_l)t+\\psi_l]\\}

    Parameters
    ----------
    Fd : float
        The Doppler frequency (in Hertz).
    Ts : float
        The sample interval (in seconds).
    L : int
        The number of rays for the Jakes model.
    shape : int | tuple[int], optional
        The shape of the sample generator. Each time the
        `generate_jakes_samples` method is called it will generate samples
        with this shape. If not provided, then 1 will be assumed. This
        could be used to generate MIMO channels. For instance, in order to
        generate channels samples for a MIMO scenario with 3 receive
        antennas and 2 transmit antennas use a shape of (3, 2).
    RS : np.random.RandomState
        The RandomState object used to generate the random values. If not
        provided, the global RandomState in numpy will be used.

    See also
    --------
    generate_jakes_samples
    """
    def __init__(self,
                 Fd: float = 100,
                 Ts: float = 1e-3,
                 L: int = 8,
                 shape: Optional[Union[int, Shape]] = None,
                 RS: Optional[np.random.RandomState] = None) -> None:
        super(JakesSampleGenerator, self).__init__(shape)

        self._Fd: float = Fd
        self._Ts: float = Ts
        self._L: int = L

        # These two will be set in the set_shape method
        self._phi_l: Optional[np.ndarray] = None
        self._psi_l: Optional[np.ndarray] = None

        if RS is None:
            # If RS was not provided, we set it to the numpy.random
            # module. That way, when the rand "method" in RS is called it
            # will actually call the global rand function in numpy.random.
            # RandomState object in numpy.
            RS = np.random
        self.RS = RS

        # self._current_time will be update after each call to the
        # `generate_more_samples` method.
        self._current_time: float = 0.0

        # Update self._phi_l and self._psi_l according to self._shape
        self._set_phi_and_psi_according_to_shape()

        # Generate first sample
        self.generate_more_samples()

    @property
    def shape(self) -> Optional[Shape]:
        """
        Get the shape of the sampling generator

        This is the shape of the samples that will be generated (not
        including num_samples).

        Returns
        -------
        tuple[int] | None
        """
        return super(JakesSampleGenerator, self).shape

    @shape.setter
    def shape(self, new_shape: Shape) -> None:
        """
        Set the shape of the sampling generator.

        This is the shape of the samples that will be generated (not
        including num_samples).

        Parameters
        ----------
        new_shape : None | int | tuple[int]
            The shape of the generated channel.
        """
        # Call the base class property setter
        FadingSampleGenerator.shape.fset(self, new_shape)  # type: ignore

        # Since phi and psi depend on the shape we need to update
        # them. Note that `_set_phi_and_psi_according_to_shape` will use
        # the new_shape of self._shape
        self._set_phi_and_psi_according_to_shape()

    @property
    def L(self) -> int:
        """The number of rays for the Jakes model"""
        return self._L

    @property
    def Ts(self) -> float:
        """The sample interval (in seconds)"""
        return self._Ts

    @property
    def Fd(self) -> float:
        """The Doppler frequency (in Hertz)"""
        return self._Fd

    def _set_phi_and_psi_according_to_shape(self) -> None:
        """
        This will update the phi and psi attributes used to generate the
        jakes samples to reflect the current value of self._shape.
        """
        if self.shape is None:
            # The dimension of phi_l and psi_l will be L x 1. We set the
            #  last dimensions as 1, instead of setting the dimension of
            #  phi_l and psi_l simply as (L,), because it will be
            # broadcasted later by numpy when we multiply with the time.
            self._phi_l = 2 * np.pi * self.RS.rand(self.L, 1)
            self._psi_l = 2 * np.pi * self.RS.rand(self.L, 1)
        else:
            # The dimension of phi_l and psi_l will be L x Shape x 1. We
            #  set the last dimensions as 1, instead of setting the
            # dimension of phi_l and psi_l simply as (L,), because it
            # will be broadcasted later by numpy when we multiply with
            # the time.
            new_shape = [self.L]
            new_shape.extend(self.shape)
            new_shape.append(1)
            self._phi_l = 2 * np.pi * self.RS.rand(*new_shape)
            self._psi_l = 2 * np.pi * self.RS.rand(*new_shape)

    def _generate_time_samples(self,
                               num_samples: Optional[int] = None
                               ) -> np.ndarray:
        """
        Generate the time samples that will be used internally in
        `generate_more_samples` method.

        Parameters
        ----------
        num_samples : int, optional
            Number of samples to be generated.

        Returns
        -------
        np.ndarray
            The numpy array with the time samples. The shape of the
            generated time variable is "(1, A, num_samples)", where 'A' has
            as many '1's as the length of self._shape.
            Ex: If self._shape is None then the shape of the returned 't'
            variable is (1, num_samples). If self._shape is (2,3) then the
            shape of the returned 't' variable is (1, 1, 1, num_samples)

        Notes
        -----
        Each time `_generate_time_samples` is called it will update
        `_current_time` to reflect the advance of the time after
        generating the new samples.
        """
        if num_samples is None:
            num_samples = 1

        # Generate a 1D numpy with the time samples
        t = np.arange(
            self._current_time,  # Start time
            num_samples * self.Ts + self._current_time,
            self.Ts * 1.0000000001)

        # Update the self._current_time variable with the value of the next
        # time sample that should be generated when _generate_time_samples
        # is called again.
        self._current_time = t[-1] + self.Ts

        # Now we will change the shape of the 't' variable to an
        # appropriated shape for later use.
        if self._shape is not None:
            # Ex: If self._shape is (2,3) then the shape of the generated
            # 't' variable should be (1,1,1,num_samples). The first
            # dimension correspond to the number of taps (that is, self.L),
            # the following two dimensions correspond to the dimensions in
            # self._shape, and the last dimension corresponds to the number
            # of time samples.
            #
            # Note that we use '1' for all dimensions except the last one
            # and numpy will replicate to the correct value later thanks to
            # broadcast.
            t.shape = [1] * (len(self._shape) + 1) + [int(num_samples)]
        else:
            # Since self._shape is None, we only need one dimension for the
            # taps (that is, self.L) and another dimension for the actual
            # time samples.
            #
            # Note that we use '1' for all dimensions except the last one
            # and numpy will replicate to the correct value later thanks to
            # broadcast.
            t.shape = (1, num_samples)

        return t

    def generate_more_samples(self, num_samples: Optional[int] = None) -> None:
        """
        Generate next samples.

        Note that any subsequent call to this method continues from the
        point where the last call stopped. That is, if you generate 10
        samples and then 15 more samples, you will get the same samples you
        would have got if you had generated 25 samples.

        Parameters
        ----------
        num_samples : int, optional
            Number of samples (with the provided shape) to generate. If not
            provided it will be assumed to be 1.

        Notes
        -----
        This method will update the self._current_time variable.
        """
        # This method will also update the _current_time member variable
        t = self._generate_time_samples(num_samples)

        # Finally calculate the channel samples
        # noinspection PyTypeChecker
        h = (math.sqrt(1.0 / self.L) * np.sum(np.exp(
            1j *
            (2 * np.pi * self.Fd * np.cos(self._phi_l) * t + self._psi_l)),
                                              axis=0))
        self._samples = h

    def skip_samples_for_next_generation(self, num_samples: int) -> None:
        """
        Advance sample generation process by `num_samples` similarly to
        what would happen if you call `generate_more_samples(
        num_samples=num_samples)`, but without actually generating the
        samples.

        This has the effect of advancing the internal time using by
        JakesSampleGenerator without generating any samples.

        Parameters
        ----------
        num_samples : int
            How many samples to skip.
        """
        self._current_time += num_samples * self.Ts

    def get_similar_fading_generator(self) -> Any:
        """
        Get a similar fading generator with the same configuration,
        but that generates independent samples.

        Returns
        -------
        JakesSampleGenerator
            Another JakesSampleGenerator object with the same configuration
            of this object.
        """
        return JakesSampleGenerator(self._Fd, self._Ts, self._L, self._shape)
